# AUTOGENERATED! DO NOT EDIT! File to edit: image-opencv_tools.ipynb (unless otherwise specified).

__all__ = ['limit_to_max_size', 'fit_to_max_size', 'get_image_size']


# Cell
import cv2


# Cell
def limit_to_max_size(img, max_size):
    """
    Limit the image size to max size and scale the image,
    if max size exceeded.
    `img`: The image to validate as OpenCV image.
    `max_size`: The max allowed image size.
    :return: The eventually resized opencv image.
    """
    biggest_size = max(img.shape[:2])
    if max_size and biggest_size > max_size:
        ratio = 1.0 * max_size / biggest_size
        img = cv2.resize(img, (int(ratio * img.shape[1]), int(ratio * img.shape[0])))
    return img


# Cell
def fit_to_max_size(img, max_width, max_height):
    """
    Limit the image size to maximum width and height and scale the image,
    if size exceeded.
    `img`: The image to validate as OpenCV Image.
    `max_width`: The max allowed image width.
    `max_height`: The max allowed image height.
    :return: The eventually resized OpenCV image.
    """
    h, w = img.shape[:2]
    scale_delta = max(w - max_width, h - max_height)
    if scale_delta > 0:
        max_size = max(w - scale_delta, h - scale_delta)
        img = limit_to_max_size(img, max_size)
    return img


# Cell
def get_image_size(fname):
    """
    Calculates image size of a given image file.
    `fname`: the file path
    return: the OpenCV image, image width and height
    """
    img = cv2.imread(fname)
    h, w = img.shape[:2]
    return img, w, h



# AUTOGENERATED! DO NOT EDIT! File to edit: io-core.ipynb (unless otherwise specified).

__all__ = ['BUF_SIZE', 'DEFAULT_IMAGE_EXTENSIONS', 'FOLDER_FILTER', 'logger', 'scan_files', 'scan_path',
           'delete_folder', 'create_folder', 'get_file_sha', 'assign_filename_suffix', 'strip_path']


# Cell
import hashlib
import logging
import shutil
import os
from os import listdir, makedirs
from os.path import join, isfile, isdir, splitext


# Cell
BUF_SIZE = 65536  # lets read stuff in 64kb chunks!
DEFAULT_IMAGE_EXTENSIONS = ['.jpg']
FOLDER_FILTER = ['.DS_Store']


# Cell
logger = logging.getLogger(__name__)


# Cell
def scan_files(folder, file_extensions=None):
    """
    Scan the folder for files and filter files by file extension.
    If the optional `file_extension` is not set, **.jpg** as file extension is used by default.
    `folder`: the folder to scan for files
    `file_extensions`: the allowed file extensions
    return: the file path list
    """
    if file_extensions is None:
        file_extensions = DEFAULT_IMAGE_EXTENSIONS

    results = []
    if isdir(folder):
        files = listdir(folder)
        for file in files:
            file_path = join(folder, file)
            filename, file_extension = splitext(file_path)
            if isfile(file_path) and (file_extension in file_extensions):
                results.append(file_path)
            elif isdir(file_path):
                results.extend(scan_files(file_path, file_extensions))
    return results


# Cell
def scan_path(path):
    """
    Scan a folder and sub-folders until the file level.
    `path`: the current directory to parse for folders
    return: a list of folders prefixed with base directory
    """
    has_sub_folders = False
    results = []
    names = listdir(path)
    for name in names:
        sub_folder_path = join(path, name)
        if isdir(sub_folder_path) and name not in FOLDER_FILTER:
            has_sub_folders = True
            results = results + scan_path(sub_folder_path)
    if not has_sub_folders:
        results.append(path)
    return results


# Cell
def delete_folder(path):
    """
    Deletes a Folder, if exist.
    `path`: the folder path including the folder name
    return: the created folder path
    """
    if isdir(path):
        shutil.rmtree(path)
    return path


# Cell
def create_folder(path, clear=False):
    """
    Creates a Folder, if not exist.
    `path`: the folder path including the folder name
    `clear`: whether or not the created folder should be empty
    return: the created folder path
    """

    if clear:
        # remove previous data-set folder, if exist to clear the contents
        delete_folder(path)

    if not isdir(path):
        makedirs(path)

    return path


# Cell
def get_file_sha(file_path):
    """
    Calculates the SHA1 of a given file.
    `file_path`: the file path
    return: the calculated SHA1 as hex
    """
    result = None
    if isfile(file_path):
        sha1 = hashlib.sha1()
        with open(file_path, 'rb') as f:
            while True:
                data = f.read(BUF_SIZE)
                if not data:
                    break
                sha1.update(data)
            result = sha1.hexdigest()
    return result


# Cell
def assign_filename_suffix(file_path, suffix=None):
    """
    Assign a filename suffix to a given file_path. (e.g <path>/filename_<suffix>.txt)
    `file_path`: the file path to suffix
    `suffix`: the suffix
    return: the suffixed file path
    """
    if suffix is None:
        return file_path
    path, ext = splitext(file_path)
    return "{}_{}{}".format(path, suffix, ext)


# Cell
def strip_path(file_path, strip=os.sep):
    """
    Strip trailing separators from a path
    `file_path`: the file path to strip
    return: the striped file path
    """
    return file_path.rstrip(os.sep)



# AUTOGENERATED! DO NOT EDIT! File to edit: annotation-folder_category_adapter.ipynb (unless otherwise specified).

__all__ = ['DEFAULT_CATEGORY_FOLDER_INDEX', 'logger', 'FolderCategoryAnnotationAdapter']


# Cell
import logging
import shutil
from os.path import join, normpath, sep, basename, isfile
from ..core import assign_arg_prefix
from ..io.core import create_folder, scan_files
from .core import Annotation, AnnotationAdapter, Region, SubsetType


# Cell
DEFAULT_CATEGORY_FOLDER_INDEX = -1


# Cell
logger = logging.getLogger(__name__)


# Cell
class FolderCategoryAnnotationAdapter(AnnotationAdapter):
    """
    Adapter to read and write annotations where the folder structure represents the categories.
    """

    def __init__(self, path, categories_file_name=None, category_index=None):
        """
        Multi Label Classification Adapter to read and write annotations.
        `path`: the folder containing the annotations
        `categories_file_name`: the name of the categories file
        `category_index`: the folder index, representing the category
        """
        super().__init__(path, categories_file_name)

        self.category_index = category_index if category_index is not None else DEFAULT_CATEGORY_FOLDER_INDEX

    @classmethod
    def argparse(cls, prefix=None):
        """
        Returns the argument parser containing argument definition for command line use.
        `prefix`: a parameter prefix to set, if needed
        return: the argument parser
        """
        parser = super(FolderCategoryAnnotationAdapter, cls).argparse(prefix=prefix)
        parser.add_argument(assign_arg_prefix('--category_index', prefix),
                            dest="category_index",
                            help="The folder index, representing the category.",
                            type=int,
                            default=None)
        return parser

    def read_annotations(self, subset_type=SubsetType.TRAINVAL):
        """
        Read annotations from folder structure representing the categories.
        `subset_type`: the subset type to read
        return: the annotations as dictionary
        """
        annotations = {}
        path = join(self.path, str(subset_type))
        logger.info('Read annotations from {}'.format(path))

        file_paths = scan_files(path)

        skipped_annotations = []
        for file_path in file_paths:
            trimmed_path = self._trim_base_path(file_path, path)
            if trimmed_path not in annotations:
                annotations[trimmed_path] = Annotation(annotation_id=trimmed_path, file_path=file_path)
            annotation = annotations[trimmed_path]

            path_split = normpath(trimmed_path).lstrip(sep).split(sep)

            if len(path_split) <= abs(self.category_index):
                logger.warning("{}: No category folder found, skip annotation.".format(trimmed_path))
                skipped_annotations.append(file_path)
                continue

            category = path_split[self.category_index - 1]
            region = Region(labels=[category])
            annotation.regions.append(region)

        logger.info('Finished read annotations')
        logger.info('Annotations read: {}'.format(len(annotations)))
        if skipped_annotations:
            logger.info('Annotations skipped: {}'.format(len(skipped_annotations)))
        return annotations

    def write_annotations(self, annotations, subset_type=SubsetType.TRAINVAL):
        """
        Write annotations to folder structure representing the categories.
        `annotations`: the annotations as dictionary
        `subset_type`: the subset type to write
        return: a list of written target file paths
        """
        path = join(self.path, str(subset_type))
        logger.info('Write annotations to {}'.format(path))
        skipped_annotations = []
        copied_files = []
        for annotation in annotations.values():
            if not isfile(annotation.file_path):
                logger.warning("{}: Source file not found, skip annotation.".format(annotation.file_path))
                skipped_annotations.append(annotation.file_path)
                continue

            skipped_labels = []
            annotation_labels = annotation.labels()
            for label in annotation_labels:
                category_folder = create_folder(join(path, label))
                target_file = join(category_folder, basename(annotation.file_path))
                if isfile(target_file):
                    logger.warning("{}: Target file already exist, skip label {}.".format(annotation.file_path, label))
                    skipped_labels.append(label)
                    continue
                # copy the file
                shutil.copy2(annotation.file_path, target_file)
                copied_files.append(target_file)
            if len(skipped_labels) == len(annotation_labels):
                logger.warning("{}: All labels skipped, skip annotation.".format(annotation.file_path))
                skipped_annotations.append(annotation.file_path)

        logger.info('Finished write annotations')
        logger.info('Annotations written: {}'.format(len(annotations) - len(skipped_annotations)))
        if skipped_annotations:
            logger.info('Annotations skipped: {}'.format(len(skipped_annotations)))
        return copied_files

    @classmethod
    def _trim_base_path(cls, file_path, base_path):
        """
        Trims the base path from a file path.
        `file_path`: the file path to trim from
        `base_path`: the base path to trim
        return: the trimmed file path
        """
        if file_path.startswith(base_path):
            file_path = file_path[len(base_path):]
        return file_path



# AUTOGENERATED! DO NOT EDIT! File to edit: tools-check_double_images.ipynb (unless otherwise specified).

__all__ = ['logger', 'BUF_SIZE', 'FILE_FILTER', 'IMAGE_EXTENSION', 'DoubleFileChecker', 'configure_logging',
           'get_file_sha', 'check_double_entries', 'remove_entries', 'scan_folder']


# Cell
import hashlib
import logging
import logging.handlers
import argparse
import sys
from os import listdir, remove
from os.path import join, isfile, isdir


# Cell
logger = logging.getLogger(__name__)


# Cell
BUF_SIZE = 65536  # lets read stuff in 64kb chunks!
FILE_FILTER = ['.DS_Store']
IMAGE_EXTENSION = '.jpg'


# Cell
class DoubleFileChecker:
    """
    Checks a given folder and subfolders for double files by calculating the corresponding SHA1.
    `path`: the folder to process
    `reverse`: if True, order the file reverse
    """

    def __init__(self, path, reverse=False):
        self.path = path
        self.reverse = reverse

    def check(self):
        """
        The main validation logic.
        """

        images = scan_folder(self.path)
        images.sort(key=len, reverse=self.reverse)
        all_images = len(images)
        contents = []
        delete_entries = []

        for index, image in enumerate(images):
            logger.info("{} / {} - Handle Image {}".format(index + 1, all_images, image))
            content = (image, get_file_sha(image))
            contents.append(content)

        # Check double entries
        logger.info('Checking Double Entries:')
        double_entries = check_double_entries(contents)

        logger.info("Found {} Entries:".format(len(double_entries)))
        for (key, entrylist) in double_entries.items():
            logger.info("{}:".format(key))
            for entry in list(entrylist):
                logger.info("-> {}".format(entry[0]))

        if len(double_entries) > 0:
            logger.info('Using only the first of each double entries.')
            delete_files = input('Will you delete the ignored double files from source? (y/n) ')
            delete_files = delete_files == 'y'
            for (key, entry) in double_entries.items():
                entry = list(entry)
                delete_entries += entry[1:]
            remove_entries(contents, delete_entries, delete_files)
        return contents, delete_entries


# Cell
def configure_logging(logging_level=logging.INFO):
    """
    Configures logging for the system.

    :param logging_level: The logging level to use.
    """
    logger.setLevel(logging_level)

    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging_level)

    logger.addHandler(handler)


# Cell
def get_file_sha(fname):
    """
    Calculates the SHA1 of a given file.
    `fname`: the file path
    return: the calculated SHA1 as hex
    """

    result = ''
    if isfile(fname):
        sha1 = hashlib.sha1()
        with open(fname, 'rb') as f:
            while True:
                data = f.read(BUF_SIZE)
                if not data:
                    break
                sha1.update(data)
            result = sha1.hexdigest()
    return result


# Cell
def check_double_entries(entries):
    """
    Process a list of tuples of filenames and corresponding hash for double hashes.
    `entries`: the list of entries with their hashes
    returns: a dictionary containing double entries by hash
    """

    hashes = dict()
    for entry in entries:
        h = entry[1]
        if h not in hashes:
            hashes[h] = []
        hashes[h].append(entry)

    double_hashes = dict()
    for (key, entrylist) in hashes.items():
        if len(entrylist) > 1:
            double_hashes[key] = entrylist
    return double_hashes


# Cell
def remove_entries(entries, to_remove, delete_source=False):
    """
    Removes entries from list and optionally remove the source file as well.
    `entries`: the list of entries to remove from
    `to_remove`: the list of entries to remove
    `delete_source`: werether or not to delete the source file as well
    returns: a list of resulting entries
    """

    for entry in to_remove:
        if entry in entries:
            index = entries.index(entry)
            logger.info("Remove Entry: ({}) {}".format(index, entry[0]))
            del entries[index]
            if delete_source:
                logger.info("Delete Source File: ".format(entry[0]))
                remove(entry[0])

    return entries


# Cell
def scan_folder(folder):
    """
    Scans a folder and subfolders for image content.
    `folder`: the folder to scan
    returns: a list of paths to images found
    """
    images = []

    contents = listdir(folder)
    for content in contents:
        if content in FILE_FILTER:
            continue
        path = join(folder, content)
        if isdir(path):
            result = scan_folder(path)
            for entry in result:
                images.append(entry)
        elif isfile(path):
            images.append(path)
    return images


# Cell
if __name__ == '__main__' and '__file__' in globals():
    configure_logging()

    parser = argparse.ArgumentParser()
    parser.add_argument("folder", help="The folder to scan.")
    parser.add_argument("--reverse-sort",
                        help="If the double entries should be sorted reverse by length.",
                        default=False,
                        action="store_true")
    args = parser.parse_args()

    checker = DoubleFileChecker(args.folder, args.reverse_sort)
    checker.check()

    logger.info('FINISHED!!!')



# AUTOGENERATED! DO NOT EDIT! File to edit: dataset-type.ipynb (unless otherwise specified).

__all__ = ['DatasetType', 'infer_dataset_type']


# Cell
from enum import Enum
from os.path import dirname, basename


# Cell
class DatasetType(Enum):
    """
    Currently supported Machine Learning Types.
    """
    IMAGE_CLASSIFICATION = 'image_classification'
    IMAGE_SEGMENTATION = 'image_segmentation'
    IMAGE_GENERATION = 'image_generation'
    IMAGE_OBJECT_DETECTION = 'image_object_detection'

    def __str__(self):
        return self.value


# Cell
def infer_dataset_type(category_file_path):
    """
    Try to infer the dataset type from the category file path.
    `category_file_path`: the path of the categories file
    return: the infered dataset type
    raises: `ValueError` if unsupported type infered
    """
    dataset_type_from_path = None
    try:
        path = dirname(category_file_path)
        dataset_type_from_path = basename(dirname(path))
        return DatasetType(dataset_type_from_path)
    except ValueError:
        raise ValueError("Error, unsupported dataset type: {}".format(dataset_type_from_path))



# AUTOGENERATED! DO NOT EDIT! File to edit: tensorflow-tfrecord_builder.ipynb (unless otherwise specified).

__all__ = ['logger', 'create_tfrecord_entry', 'create_tfrecord_file', 'int64_feature', 'int64_list_feature',
           'bytes_feature', 'bytes_list_feature', 'float_list_feature', 'create_labelmap_file']


# Cell
import logging
import io
import tensorflow as tf
from os import environ
from os.path import basename
from ..image.pillow_tools import get_image_size


# Cell
environ['TF_CPP_MIN_LOG_LEVEL'] = '2'    # Suppress TensorFlow logging (1)

logger = logging.getLogger(__name__)


# Cell
def create_tfrecord_entry(categories, annotation):
    """
    Create tfrecord entry with annotations for one file / image.
    `categories`: the categories used
    `annotation`: the annotation of a file / image
    return: the tfrecord entry
    """
    with tf.io.gfile.GFile(annotation.file_path, 'rb') as fid:
        encoded_jpg = fid.read()
    _, width, height = get_image_size(io.BytesIO(encoded_jpg))

    file_name = basename(annotation.file_path).encode('utf8')
    image_format = b'jpg'
    xmins = []
    xmaxs = []
    ymins = []
    ymaxs = []
    classes_text = []
    classes = []

    for region in annotation.regions:
        x_min = min(region.points_x)
        y_min = min(region.points_y)
        x_max = max(region.points_x)
        y_max = max(region.points_y)
        category = region.labels[0] if len(region.labels) else ''

        xmins.append(x_min / width)
        xmaxs.append(x_max / width)
        ymins.append(y_min / height)
        ymaxs.append(y_max / height)
        classes_text.append(category.encode('utf8'))
        classes.append(categories.index(category))

    tf_example = tf.train.Example(features=tf.train.Features(feature={
        'image/height': int64_feature(height),
        'image/width': int64_feature(width),
        'image/filename': bytes_feature(file_name),
        'image/source_id': bytes_feature(file_name),
        'image/encoded': bytes_feature(encoded_jpg),
        'image/format': bytes_feature(image_format),
        'image/object/bbox/xmin': float_list_feature(xmins),
        'image/object/bbox/xmax': float_list_feature(xmaxs),
        'image/object/bbox/ymin': float_list_feature(ymins),
        'image/object/bbox/ymax': float_list_feature(ymaxs),
        'image/object/class/text': bytes_list_feature(classes_text),
        'image/object/class/label': int64_list_feature(classes),
    }))
    return tf_example


# Cell
def create_tfrecord_file(output_path, categories, annotations):
    """
    Create a tfrecord file for a sub-data-set, which can be one of the following: training, validation, test
    `output_path`: the path including the filename of the tfrecord file
    `categories`: the categories used
    `annotations`: the annotations of the files / images
    """
    writer = tf.io.TFRecordWriter(output_path)
    for annotation in annotations.values():
        tf_example = create_tfrecord_entry(categories, annotation)
        writer.write(tf_example.SerializeToString())
    writer.close()
    logger.info('Successfully created the TFRecord file: {}'.format(output_path))


# Cell
def int64_feature(value):
    return tf.train.Feature(int64_list=tf.train.Int64List(value=[value]))


def int64_list_feature(value):
    return tf.train.Feature(int64_list=tf.train.Int64List(value=value))


def bytes_feature(value):
    return tf.train.Feature(bytes_list=tf.train.BytesList(value=[value]))


def bytes_list_feature(value):
    return tf.train.Feature(bytes_list=tf.train.BytesList(value=value))


def float_list_feature(value):
    return tf.train.Feature(float_list=tf.train.FloatList(value=value))


# Cell
def create_labelmap_file(output_path, categories, start=1):
    """
    Create labelmap protobuffer text file containing the categories.
    Format is compatible with Tensorflow Object Detection API.
    For object detection data-sets, the categories should exclude the background class and `start` should be 1.
    `output_path`: the path including the filename of the protobuffer text file
    `categories`: a list of the categories to write
    `start`: the category index for the first category
    """
    # create label_map data
    label_map = ''
    for index, category in enumerate(categories, start=start):
        label_map = label_map + "item {\n"
        label_map = label_map + " id: " + str(index) + "\n"
        label_map = label_map + " name: '" + category + "'\n}\n\n"
    label_map = label_map[:-1]

    # write label_map file
    with open(output_path, "w") as f:
        f.write(label_map)
        f.close()



# AUTOGENERATED! DO NOT EDIT! File to edit: dataset-core.ipynb (unless otherwise specified).

__all__ = ['logger', 'Dataset']


# Cell
import argparse
import logging
from abc import ABC
from os.path import join, basename, dirname
from ..core import assign_arg_prefix
from ..annotation.core import AnnotationAdapter, SubsetType
from ..image.pillow_tools import assign_exif_orientation, write_exif_metadata
from ..io.core import create_folder


# Cell
logger = logging.getLogger(__name__)


# Cell
class Dataset(ABC):
    """
    Dataset base class to build datasets.
    `args`: the arguments containing the parameters
    """

    DEFAULT_SPLIT = 0.2

    def __init__(self, input_adapter: AnnotationAdapter, output_adapter: AnnotationAdapter, split=None, seed=None,
                 sample=None):
        self.input_adapter = input_adapter
        self.output_adapter = output_adapter
        self.split = self.DEFAULT_SPLIT if split is None else split
        self.seed = seed
        self.sample = sample
        self.categories = input_adapter.read_categories()
        self.annotations = input_adapter.read_annotations()

    @classmethod
    def argparse(cls, prefix=None):
        """
        Returns the argument parser containing argument definition for command line use.
        `prefix`: a parameter prefix to set, if needed
        return: the argument parser
        """
        parser = argparse.ArgumentParser()
        parser.add_argument(assign_arg_prefix('--split', prefix),
                            dest="split",
                            help="Percentage of the data which belongs to validation set.",
                            type=float,
                            default=0.2)
        parser.add_argument(assign_arg_prefix('--seed', prefix),
                            dest="seed",
                            help="A random seed to reproduce splits.",
                            type=int,
                            default=None)
        parser.add_argument(assign_arg_prefix('--sample', prefix),
                            dest="sample",
                            help="Percentage of the data which will be copied as a sample set.",
                            type=float,
                            default=0)

        return parser

    def create_folders(self):
        """
        Creates the dataset folder structure, if not exist
        """
        output_folder = create_folder(self.output_adapter.path, clear=True)
        logger.info("Created folder {}".format(output_folder))

    def build_info(self):
        """
        Log build information
        """
        logger.info('Build configuration:')
        logger.info('input_adapter: {}'.format(type(self.input_adapter).__name__))
        logger.info('input_path: {}'.format(self.input_adapter.path))
        logger.info('output_adapter: {}'.format(type(self.output_adapter).__name__))
        logger.info('output_path: {}'.format(self.output_adapter.path))
        logger.info('split: {}'.format(self.split))
        logger.info('seed: {}'.format(self.seed))
        logger.info('sample: {}'.format(self.sample))

    def validate(self):
        """
        Validates the annotations.
        return: The skipped annotations
        """
        # validate only the trainval images, the test images have no annotations to validate
        logger.info('Start validate data at {}'.format(self.input_adapter.path))

        files = self.input_adapter.list_files()

        logger.info('Found {} files at {}'.format(len(files), self.input_adapter.path))

        delete_annotations = {}
        used_categories = set([])

        for annotation_id, annotation in self.annotations.items():

            delete_regions = {}
            for index, region in enumerate(annotation.regions):
                len_labels = len(region.labels)
                region_valid = len_labels and len(set(region.labels) & set(self.categories)) == len_labels
                if not region_valid:
                    message = '{} : Region {} with category {} is not in category list, skip region.'
                    logger.info(message.format(annotation.file_path, index, ','.join(region.labels)))

                    delete_regions[index] = True
                else:
                    # update the used regions
                    used_categories.update(region.labels)

            # delete regions after iteration is finished
            for index in sorted(list(delete_regions.keys()), reverse=True):
                del annotation.regions[index]

            # validate for empty region
            if not annotation.regions:
                logger.info('{} : Has empty regions, skip annotation.'.format(annotation.file_path))
                delete_annotations[annotation_id] = True
            # validate for file exist
            elif annotation.file_path not in files:
                logger.info('{} : File of annotations do not exist, skip annotations.'.format(annotation.file_path))
                delete_annotations[annotation_id] = True
            else:
                files.pop(files.index(annotation.file_path))

        for index, file in enumerate(files):
            logger.info('[{}] -> {} : File has no annotations, skip file.'.format(index, file))

        # list unused categories
        empty_categories = frozenset(self.categories) - used_categories
        if empty_categories:
            logger.info('The following categories have no images: {}'.format(" , ".join(empty_categories)))

        # delete annotations after iteration is finished
        for index in delete_annotations.keys():
            del self.annotations[index]

        logger.info('Finished validate image set at {}'.format(self.input_adapter.path))
        return delete_annotations

    def copy(self, train_annotation_keys, val_annotation_keys, test_files=None):
        """
        Copy the images to the dataset.
        `train_annotation_keys`: The list of training annotation keys
        `val_annotation_keys`: The list of validation annotation keys
        `test_files`: The list of test file paths
        return: A tuple containing train, val and test target file paths
        """

        logger.info('Start copy annotations from {} to {}'.format(self.input_adapter.path,
                                                                  self.output_adapter.path))

        # copy the categories files
        logger.info('Write categories to {}'.format(self.output_adapter.path))
        self.output_adapter.write_categories(self.categories)

        logger.info('Write {} annotations to {}'.format(str(SubsetType.TRAIN), self.output_adapter.path))
        annotations_train = dict(zip(train_annotation_keys, [self.annotations[key] for key in train_annotation_keys]))
        train_targets = self.output_adapter.write_annotations(annotations_train, SubsetType.TRAIN)
        logger.info('Write {} annotations to {}'.format(str(SubsetType.VAL), self.output_adapter.path))
        annotations_val = dict(zip(val_annotation_keys, [self.annotations[key] for key in val_annotation_keys]))
        val_targets = self.output_adapter.write_annotations(annotations_val, SubsetType.VAL)
        logger.info('Write {} files to {}'.format(str(SubsetType.TEST), self.output_adapter.path))
        test_targets = self.output_adapter.write_files(test_files, SubsetType.TEST) if test_files else []

        return train_targets, val_targets, test_targets

    def build(self, validate=True):
        """
        Build the data-set. This is the main logic.
        This method validates the images against the annotations,
        split the image-set into train and val on given split percentage,
        creates the data-set folders and copies the image.
        If a sample percentage is given, a sub-set is created as sample.
        `validate`: True if annotations should be validate, else False
        """
        logger.info('Validation set contains {}% of the images.'.format(int(self.split * 100)))

        # validate the image set
        skipped_annotations = self.validate() if validate else {}

        # split category files into train & val and create the sample split, if set
        train_annotation_keys = []
        val_annotation_keys = []
        sample_train_annotation_keys = []
        sample_val_annotation_keys = []

        if self.split == 0:
            train, val = (list(self.annotations.keys()), [])
        elif self.split == 1:
            train, val = ([], list(self.annotations.keys()))
        else:
            train, val = self.split_train_val_data(list(self.annotations.keys()), self.split, self.seed)
        train_annotation_keys.extend(train)
        val_annotation_keys.extend(val)

        # if a sample data set should be created, create the splits
        if self.sample:
            _, sample_train = self.split_train_val_data(train, self.sample, self.seed)
            _, sample_val = self.split_train_val_data(val, self.sample, self.seed)
            sample_train_annotation_keys.extend(sample_train)
            sample_val_annotation_keys.extend(sample_val)

        # if test files exist
        test_files = self.input_adapter.list_files(SubsetType.TEST)
        if test_files and self.sample:
            _, sample_test_files = self.split_train_val_data(test_files, self.sample, self.seed)
        else:
            sample_test_files = None

        # copy the annotations
        self.copy(train_annotation_keys, val_annotation_keys, test_files)

        if self.sample:
            # backup original output path
            output_path = self.output_adapter.path
            sample_name = "{}_sample".format(basename(output_path))
            # set output path to sample set
            self.output_adapter.path = join(dirname(output_path), sample_name)
            logger.info('Start build {} dataset containing {}% of images at {}'.format(sample_name,
                                                                                       int(self.sample * 100),
                                                                                       self.output_adapter.path))
            # create the sample data set folder
            create_folder(self.output_adapter.path)
            # copy the sample data
            self.copy(sample_train_annotation_keys, sample_val_annotation_keys, sample_test_files)

            logger.info('Finished build {} dataset containing {}% of images at {}'.format(sample_name,
                                                                                          int(self.sample * 100),
                                                                                          self.output_adapter.path))
            # restore original output path
            self.output_adapter.path = output_path

    @classmethod
    def split_train_val_data(cls, data, val_size=0.2, seed=None):
        """
        Splits the images in train and validation set
        `data`: the data to split
        `val_size`: the size of the validation set in percentage
        `seed`: A random seed to reproduce splits.
        return: the split train, validation images
        """
        from sklearn.model_selection import train_test_split
        train, test = train_test_split(data, random_state=seed, test_size=val_size) if len(data) > 1 else (data, [])
        return train, test

    @classmethod
    def assign_orientation(cls, file_path):
        """
        Assign the EXIF metadata orientation to an image.
        `file_path`: the path to the image file
        """

        # rotate image by EXIF orientation metadata and remove them
        image, exif_data, rotated = assign_exif_orientation(file_path)
        if rotated:
            write_exif_metadata(image, exif_data, file_path)



# AUTOGENERATED! DO NOT EDIT! File to edit: annotation-core.ipynb (unless otherwise specified).

__all__ = ['logger', 'SubsetType', 'RegionShape', 'parse_region_shape', 'Region', 'Annotation', 'convert_region',
           'region_bounding_box', 'AnnotationAdapter']


# Cell
import argparse
import logging
import shutil
from abc import ABC, abstractmethod
from enum import Enum
from copy import deepcopy
from os.path import join, isdir, basename
from aiforce import category_tools
from ..core import assign_arg_prefix
from ..io.core import strip_path, scan_files, create_folder


# Cell
logger = logging.getLogger(__name__)


# Cell
class SubsetType(Enum):
    """
    The annotation subset type
    """
    TRAINVAL = 'trainval'
    TRAIN = 'train'
    VAL = 'val'
    TEST = 'test'

    def __str__(self):
        return self.value


# Cell
class RegionShape(Enum):
    """
    The supported region shape types
    """
    NONE = 'none'
    CIRCLE = 'circle'
    ELLIPSE = 'ellipse'
    POINT = 'point'
    POLYGON = 'polygon'
    RECTANGLE = 'rect'

    def __str__(self):
        return self.value


# Cell
def parse_region_shape(shape_str):
    """
    Try to parse the region shape from a string representation.
    `shape_str`: the shape as string
    return: the parsed RegionShape
    raises: `ValueError` if unsupported shape parsed
    """
    try:
        return RegionShape(shape_str)
    except ValueError:
        raise ValueError("Error, unsupported region shape: {}".format(shape_str))


# Cell
class Region:
    """
    A region
    `shape`: the region shape
    `points_x`: a list of points x-coordinates
    `points_y`: a list of points y-coordinates
    `radius_x`: a radius on x-coordinate
    `radius_y`: a radius on y-coordinate
    `labels`: a set of region labels
    """

    def __init__(self, shape=RegionShape.NONE, points_x=None, points_y=None, radius_x=0, radius_y=0, labels=None):
        self.shape = shape
        self.points_x = [] if points_x is None else points_x
        self.points_y = [] if points_y is None else points_y
        self.radius_x = radius_x
        self.radius_y = radius_y
        self.labels = [] if labels is None else labels


# Cell
class Annotation:
    """
    A annotation for a file.
    `annotation_id`: a unique annotation identifier
    `file_name`: the file
    `file_path`: the file path
    `regions`: A list of regions
    """

    def __init__(self, annotation_id=None, file_path=None, regions=None):
        self.annotation_id = annotation_id
        self.file_path = file_path
        self.regions: [Region] = [] if regions is None else regions

    def labels(self):
        """
        Returns a list of labels, assigned to the annotation.
        return: a list of labels
        """
        labels = {}
        for region in self.regions:
            for label in region.labels:
                labels[label] = None
        return list(labels.keys())


# Cell
def convert_region(region: Region, target_shape: RegionShape):
    """
    Convert region to target shape.
    `region`: the region to convert
    `target_shape`: the target shape to convert to
    """
    if target_shape != region.shape:
        x_min = min(region.points_x) - region.radius_x if len(region.points_x) else 0
        x_max = max(region.points_x) + region.radius_x if len(region.points_x) else 0
        y_min = min(region.points_y) - region.radius_y if len(region.points_y) else 0
        y_max = max(region.points_y) + region.radius_y if len(region.points_y) else 0
        center_x = x_min + x_max - x_min
        center_y = y_min + y_max - y_min
        region.shape = target_shape
        if target_shape == RegionShape.NONE:
            region.points_x = []
            region.points_y = []
            region.radius_x = 0
            region.radius_y = 0
        elif target_shape == RegionShape.CIRCLE or target_shape == RegionShape.ELLIPSE:
            region.points_x = [center_x]
            region.points_y = [center_y]
            region.radius_x = x_max - center_x
            region.radius_y = y_max - center_y
        elif target_shape == RegionShape.POINT:
            region.points_x = [center_x]
            region.points_y = [center_y]
            region.radius_x = 0
            region.radius_y = 0
        elif target_shape == RegionShape.POLYGON:
            region.points_x = [x_min, x_min, x_max, x_max, x_min]
            region.points_y = [y_min, y_max, y_max, y_min, y_min]
            region.radius_x = 0
            region.radius_y = 0
        elif target_shape == RegionShape.RECTANGLE:
            region.points_x = [x_min, x_max]
            region.points_y = [y_min, y_max]
            region.radius_x = 0
            region.radius_y = 0
        else:
            raise NotImplementedError('unsupported conversion {} -> {}'.format(region.shape, target_shape))


# Cell
def region_bounding_box(region: Region):
    """
    Calculates the region bounding box.
    `region`: the region
    return: a tuple of points_x and points_y
    """
    bbox = deepcopy(region)
    convert_region(bbox, RegionShape.RECTANGLE)
    return bbox.points_x, bbox.points_y


# Cell
class AnnotationAdapter(ABC):
    """
    Abstract Base Adapter to inherit for writing custom adapters
    """
    DEFAULT_CATEGORIES_FILE = 'categories.txt'

    def __init__(self, path, categories_file_name=None):
        """
        Base Adapter to read and write annotations.
        `path`: the folder containing the annotations
        `categories_file_name`: the name of the categories file
        """
        self.path = strip_path(path)
        if categories_file_name is None:
            self.categories_file_name = self.DEFAULT_CATEGORIES_FILE
        else:
            self.categories_file_name = categories_file_name

    def list_files(self, subset_type=SubsetType.TRAINVAL):
        """
        List all physical files in a sub-set.
        `subset_type`: the subset type to list
        return: a list of file paths if subset type exist, else an empty list
        """
        path = join(self.path, str(subset_type))
        return scan_files(path) if isdir(path) else []

    def write_files(self, file_paths, subset_type=SubsetType.TRAINVAL):
        """
        Write physical files in a sub-set.
        `file_paths`: a list of file paths to write
        `subset_type`: the subset type to write into
        return: a list of written target file paths
        """
        path = join(self.path, str(subset_type))
        create_folder(path)
        copied_files = []
        for file_path in file_paths:
            file_name = basename(file_path)
            target_file = join(path, file_name)
            shutil.copy2(file_path, target_file)
            copied_files.append(target_file)
        return copied_files

    @abstractmethod
    def read_annotations(self, subset_type=SubsetType.TRAINVAL):
        """
        Read annotations.
        `subset_type`: the subset type to read
        return: the annotations as dictionary
        """
        pass

    def read_categories(self):
        """
        Read categories.
        return: a list of category names
        """
        path = join(self.path, self.categories_file_name)
        logger.info('Read categories from {}'.format(path))
        return category_tools.read_categories(path)

    @abstractmethod
    def write_annotations(self, annotations, subset_type=SubsetType.TRAINVAL):
        """
        Write annotations.
        `annotations`: the annotations as dictionary
        `subset_type`: the subset type to write
        return: a list of written target file paths
        """
        pass

    def write_categories(self, categories):
        """
        Write categories.
        `categories`: a list of category names
        """
        target_folder = create_folder(self.path)
        path = join(target_folder, self.categories_file_name)
        logger.info('Write categories to {}'.format(path))
        category_tools.write_categories(categories, path)

    @classmethod
    def argparse(cls, prefix=None):
        """
        Returns the argument parser containing argument definition for command line use.
        `prefix`: a parameter prefix to set, if needed
        return: the argument parser
        """
        parser = argparse.ArgumentParser()
        parser.add_argument(assign_arg_prefix('--path', prefix),
                            dest="path",
                            help="the folder containing the annotations.",
                            required=True)
        parser.add_argument(assign_arg_prefix('--categories_file_name', prefix),
                            dest="categories_file_name",
                            help="The name of the categories file.",
                            default=None)
        return parser



# AUTOGENERATED! DO NOT EDIT! File to edit: image-pillow_tools.ipynb (unless otherwise specified).

__all__ = ['EXIF_ORIENTATION_TAG', 'logger', 'limit_to_max_size', 'fit_to_max_size', 'get_image_size',
           'get_image_orientation', 'read_exif_metadata', 'write_exif_metadata', 'assign_exif_orientation',
           'convert_to_base64', 'write_mask', 'configure_logging']


# Cell
import sys
import argparse
import logging
import base64
import piexif
from io import BytesIO
from PIL import Image as PILImage
from .tools import ImageOrientation
from ..io.core import scan_files


# Cell
EXIF_ORIENTATION_TAG = 'Orientation'
"""The Image EXIF orientation tag"""


# Cell
logger = logging.getLogger(__name__)


# Cell
def limit_to_max_size(img, max_size):
    """
    Limit the image size to max size and scale the image,
    if max size exceeded.
    `img`: The image to validate as Pillow Image.
    `max_size`: The max allowed image size.
    :return: The eventually resized image.
    """
    biggest_size = max(img.size)
    if max_size and biggest_size > max_size:
        ratio = 1.0 * max_size / biggest_size
        img = img.resize([int(ratio * s) for s in img.size])
    return img


# Cell
def fit_to_max_size(img, max_width, max_height):
    """
    Limit the image size to maximum width and height and scale the image,
    if size exceeded.
    `img`: The image to validate as Pillow Image.
    `max_width`: The max allowed image width.
    `max_height`: The max allowed image height.
    :return: The eventually resized image.
    """
    w, h = img.size
    scale_delta = max(w - max_width, h - max_height)
    if scale_delta > 0:
        max_size = max(w - scale_delta, h - scale_delta)
        img = limit_to_max_size(img, max_size)
    return img


# Cell
def get_image_size(fname):
    """
    Calculates image size of a given image file path.
    `fname`: the file path
    return: the Pillow image, image width and height
    """
    image = PILImage.open(fname)
    w, h = image.size
    return image, w, h


# Cell
def get_image_orientation(fname):
    """
    Parses the EXIF orientation information from the image.
    `fname`: the file path
    :return: The Pillow image and the orientation of the image.
    """
    orientation = ImageOrientation.TOP
    image, exif_data = read_exif_metadata(fname)
    if exif_data is not None and "0th" in exif_data:
        exif_data_0 = exif_data["0th"]
        if piexif.ImageIFD.Orientation in exif_data_0:
            try:
                orientation = ImageOrientation(exif_data_0[piexif.ImageIFD.Orientation])
            except ValueError as e:
                logger.error(e)
    return image, orientation


# Cell
def read_exif_metadata(fname):
    """
    Read the EXIF metadata information from the image.
    `fname`: the file path
    :return: The Pillow image, EXIF metadata as dictionary or None, if no EXIF data exist.
    """
    image = PILImage.open(fname)
    exif_data = None
    if "exif" in image.info:
        exif_data = piexif.load(image.info["exif"])
    return image, exif_data


# Cell
def write_exif_metadata(image, exif_data, fname):
    """
    Write the EXIF metadata information to the image.
    `image`: the Pillow image to write the EXIF metadata to
    `exif_data`: the EXIF metadata as dictionary
    `fname`: a file path to store the image
    :return: `True` if EXIF metadata saved, else `False`
    """
    if image and exif_data:
        piexif.dump(exif_data)
        exif_bytes = piexif.dump(exif_data)
        image.save(fname, exif=exif_bytes)
        return True
    return False


# Cell
def assign_exif_orientation(fname):
    """
    Parses the EXIF orientation metadata from the image,
    rotate the image accordingly and remove the image EXIF orientation metadata.
    `fname`: the file path
    :return: The eventually rotated Pillow image, EXIF metadata with orientation information stripped
    and `True` if image rotated, else `False`
    """
    image, exif_data = read_exif_metadata(fname)

    orientation = ImageOrientation.TOP

    if exif_data and piexif.ImageIFD.Orientation in exif_data["0th"]:
        orientation = exif_data["0th"].pop(piexif.ImageIFD.Orientation)
        try:
            orientation = ImageOrientation(orientation)
        except ValueError as e:
            logger.error(e)

        if orientation == ImageOrientation.TOP_FLIPPED:
            image = image.transpose(PILImage.FLIP_LEFT_RIGHT)
        elif orientation == ImageOrientation.BOTTOM:
            image = image.rotate(180)
        elif orientation == ImageOrientation.BOTTOM_FLIPPED:
            image = image.rotate(180).transpose(PILImage.FLIP_LEFT_RIGHT)
        elif orientation == ImageOrientation.RIGHT_FLIPPED:
            image = image.rotate(-90, expand=True).transpose(PILImage.FLIP_LEFT_RIGHT)
        elif orientation == ImageOrientation.RIGHT:
            image = image.rotate(-90, expand=True)
        elif orientation == ImageOrientation.LEFT_FLIPPED:
            image = image.rotate(90, expand=True).transpose(PILImage.FLIP_LEFT_RIGHT)
        elif orientation == ImageOrientation.LEFT:
            image = image.rotate(90, expand=True)

    rotated = orientation != ImageOrientation.TOP

    return image, exif_data, rotated


# Cell
def convert_to_base64(image, image_type="PNG"):
    """
    Converts the specified image into a base64 version of itself.

    `image`: The image to transform as Pillow Image.
    `image_type`: The image type.
    :return: The base64 encoded version of the image.
    """
    buffered = BytesIO()
    image.save(buffered, format=image_type)
    return base64.b64encode(buffered.getvalue()).decode('UTF-8')


# Cell
def write_mask(mask, file_path, palette=None):
    """
    Writes a mask as image.
    The image format is set by the file extension in file_path.

    `mask`: The mask to write.
    `file_path`: The path including file name and file extension to write to.
    `palette`: The color palette to assign, if set.
    """
    im = PILImage.fromarray(mask)
    if palette is not None:
        im.putpalette(palette)
    im.save(file_path)


# Cell
def configure_logging(logging_level=logging.INFO):
    """
    Configures logging for the system.

    :param logging_level: The logging level to use.
    """
    logger.setLevel(logging_level)

    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging_level)

    logger.addHandler(handler)


# Cell
if __name__ == '__main__' and '__file__' in globals():
    # for direct shell execution
    configure_logging()

    parser = argparse.ArgumentParser()
    parser.add_argument("image_path",
                        help="The path to the image files.")

    args = parser.parse_args()
    files = scan_files(args.image_path)
    for file in files:
        _, w, h = get_image_size(file)
        _, orientation = get_image_orientation(file)
        logger.info("Size: width: {}, height: {}, orientation: {}".format(w, h, orientation))



# AUTOGENERATED! DO NOT EDIT! File to edit: annotation-multi_category_adapter.ipynb (unless otherwise specified).

__all__ = ['DEFAULT_ANNOTATIONS_FILE', 'CSV_FIELDNAMES', 'logger', 'MultiCategoryAnnotationAdapter']


# Cell
import csv
import shutil
import logging
from os.path import join, basename, isfile, splitext
from ..core import assign_arg_prefix
from ..io.core import create_folder
from .core import Annotation, AnnotationAdapter, Region, SubsetType


# Cell
DEFAULT_ANNOTATIONS_FILE = 'annotations.csv'
CSV_FIELDNAMES = ['image_name', 'tags']


# Cell
logger = logging.getLogger(__name__)


# Cell
class MultiCategoryAnnotationAdapter(AnnotationAdapter):
    """
    Adapter to read and write annotations for multi label classification.
    """

    def __init__(self, path, categories_file_name=None, annotations_file_name=None):
        """
        Multi Label Classification Adapter to read and write annotations.
        `path`: the folder containing the annotations
        `categories_file_name`: the name of the categories file
        `annotations_file_name`: the name of annotations file
        """
        super().__init__(path, categories_file_name)

        if annotations_file_name is None:
            self.annotations_file_name = DEFAULT_ANNOTATIONS_FILE
        else:
            self.annotations_file_name = annotations_file_name

    @classmethod
    def argparse(cls, prefix=None):
        """
        Returns the argument parser containing argument definition for command line use.
        `prefix`: a parameter prefix to set, if needed
        return: the argument parser
        """
        parser = super(MultiCategoryAnnotationAdapter, cls).argparse(prefix=prefix)
        parser.add_argument(assign_arg_prefix('--annotations_file', prefix),
                            dest="annotations_file_name",
                            help="The name of the multi classification CSV annotation file.",
                            default=None)
        return parser

    def read_annotations(self, subset_type=SubsetType.TRAINVAL):
        """
        Read annotations from a multi classification CSV annotations file.
        `subset_type`: the subset type to read
        return: the annotations as dictionary
        """
        path = join(self.path, str(subset_type))
        annotations_file_name = self._annotation_file_name_suffix_handling(subset_type)
        annotations_file_path = join(self.path, annotations_file_name)
        logger.info('Read file sources from {}'.format(path))
        logger.info('Read annotations from {}'.format(annotations_file_path))

        annotations = {}

        with open(annotations_file_path, newline='') as csv_file:
            reader = csv.DictReader(csv_file)

            skipped_annotations = []
            for row in reader:
                file_path = join(path, row['image_name'])
                if not isfile(file_path):
                    logger.warning("{}: Source file not found, skip annotation.".format(file_path))
                    skipped_annotations.append(file_path)
                    continue

                if file_path not in annotations:
                    annotations[file_path] = Annotation(annotation_id=file_path, file_path=file_path)

                annotation = annotations[file_path]

                tags = row['tags'] if 'tags' in row else []
                for category in tags.split(' '):
                    region = Region(labels=[category])
                    annotation.regions.append(region)

        logger.info('Finished read annotations')
        logger.info('Annotations read: {}'.format(len(annotations)))
        if skipped_annotations:
            logger.info('Annotations skipped: {}'.format(len(skipped_annotations)))
        return annotations

    def write_annotations(self, annotations, subset_type=SubsetType.TRAINVAL):
        """
        Writes a multi classification CSV annotations file and copy the corresponding source files.
        `annotations`: the annotations as dictionary
        `subset_type`: the subset type to write
        return: a list of written target file paths
        """
        path = join(self.path, str(subset_type))
        target_folder = create_folder(path)
        annotations_file_name = self._annotation_file_name_suffix_handling(subset_type)
        annotations_file_path = join(self.path, annotations_file_name)
        logger.info('Write file sources to {}'.format(target_folder))
        logger.info('Write annotations to {}'.format(annotations_file_path))

        skipped_annotations = []
        copied_files = []
        rows = {}
        for annotation in annotations.values():
            target_file_name = basename(annotation.file_path)
            target_file = join(target_folder, target_file_name)

            if not isfile(annotation.file_path):
                logger.warning("{}: Source file not found, skip annotation.".format(annotation.file_path))
                skipped_annotations.append(annotation.file_path)
                continue

            # copy the file
            if target_file_name not in rows:
                rows[target_file_name] = []
            rows[target_file_name].extend(annotation.labels())

            if isfile(target_file):
                logger.warning("{}: Target file already exist, skip annotation.".format(annotation.file_path))
                skipped_annotations.append(annotation.file_path)
                continue
            shutil.copy2(annotation.file_path, target_file)
            copied_files.append(target_file)
        with open(annotations_file_path, 'w', newline='') as csv_file:
            writer = csv.DictWriter(csv_file, fieldnames=CSV_FIELDNAMES, delimiter=',', quotechar='"',
                                    quoting=csv.QUOTE_MINIMAL)
            writer.writeheader()
            writer.writerows([dict(zip(CSV_FIELDNAMES, [key, ' '.join(labels)])) for key, labels in rows.items()])

        logger.info('Finished write annotations')
        logger.info('Annotations written: {}'.format(len(annotations) - len(skipped_annotations)))
        if skipped_annotations:
            logger.info('Annotations skipped: {}'.format(len(skipped_annotations)))
        return copied_files

    def _annotation_file_name_suffix_handling(self, subset_type):
        """
        Handle annotations file name based on the subset type.
        `subset_type`: the subset type to handle
        return: the annotations file name
        """
        file_name, ext = splitext(self.annotations_file_name)
        if subset_type in [SubsetType.TRAIN, SubsetType.VAL] and not file_name.endswith(str(subset_type)):
            suffix = "_{}".format(str(subset_type))
            return "{}{}{}".format(file_name, suffix, ext)
        return self.annotations_file_name



# AUTOGENERATED! DO NOT EDIT! File to edit: tools-image_size_calculator.ipynb (unless otherwise specified).

__all__ = ['logger', 'ImageSizeCalculator', 'configure_logging']


# Cell
import logging
import logging.handlers
import argparse
import sys
from functools import reduce
from ..image.pillow_tools import get_image_size
from ..io.core import scan_path


# Cell
logger = logging.getLogger(__name__)


# Cell
class ImageSizeCalculator:
    """
    Calculates image sizes in a given folder and subfolders. Summarize unique image sizes at the end.
    `path`: the folder to process
    """

    def __init__(self, path):
        self.path = path

    def calculate(self):
        """
        The main logic.
        return: a dictionary with unique sizes as key and image count as value
        """

        images = scan_path(self.path)
        all_images = len(images)
        unique_sizes = {}

        for index, image in enumerate(images):
            _, width, height = get_image_size(image)
            size_key = "{}x{}".format(width, height)
            logger.info("{} / {} - Handle Image {} with size {}x{}".format(
                index + 1,
                all_images,
                image,
                width,
                height,
            ))
            if size_key not in unique_sizes:
                unique_sizes[size_key] = 0
            unique_sizes[size_key] += 1

        return unique_sizes


# Cell
def configure_logging(logging_level=logging.INFO):
    """
    Configures logging for the system.

    :param logging_level: The logging level to use.
    """
    logger.setLevel(logging_level)

    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging_level)

    logger.addHandler(handler)


# Cell
if __name__ == '__main__' and '__file__' in globals():
    configure_logging()

    parser = argparse.ArgumentParser()
    parser.add_argument("folder", help="The folder to scan.")
    args = parser.parse_args()

    calculator = ImageSizeCalculator(args.folder)
    sizes = calculator.calculate()
    image_count = reduce(lambda a, x: a + x, sizes.values())
    logger.info("Images Analyzed: {}".format(image_count))

    logger.info("Unique Image Sizes and Image Count:")
    for key, size in sizes.items():
        logger.info("{}: {}".format(key, size))

    logger.info('FINISHED!!!')



# AUTOGENERATED! DO NOT EDIT! File to edit: core.ipynb (unless otherwise specified).

__all__ = ['DEFAULT_HELP_ARGUMENTS', 'import_modules', 'list_subclasses', 'all_subclasses', 'parse_known_help',
           'parse_known_args_with_help', 'assign_arg_prefix', 'input_feedback']


# Cell
import re
from importlib import import_module, resources


# Cell
DEFAULT_HELP_ARGUMENTS = ['-h', '--help']


# Cell
def import_modules(package):
    """
    Dynamically import all modules but not submodules in a package.
    `package`: the package to import all modules
    """
    files = resources.contents(package)
    modules = [f[:-3] for f in files if f.endswith(".py") and f[0] != "_"]
    for module in modules:
        import_module(f"{package.__name__}.{module}")


# Cell
def list_subclasses(package, base_class):
    """
    Dynamically import all modules in a package and scan for all subclasses of a base class.
    `package`: the package to import
    `base_class`: the base class to scan for subclasses
    return: a dictionary of possible subclasses with class name as key and class type information as value
    """
    import_modules(package)
    subclasses = all_subclasses(base_class)
    return dict(zip(map(lambda c: c.__name__, subclasses), subclasses))


# Cell
def all_subclasses(base_class):
    """
    Scan recursively for all subclasses of a base class.
    `base_class`: the base class to scan for subclasses
    return: a set of subclasses type information
    """
    return set(base_class.__subclasses__()).union([s for c in base_class.__subclasses__() for s in all_subclasses(c)])


# Cell
def parse_known_help(argv):
    """
    Parse the command line arguments for help parameter and decide,
    if the help parameter should be assigned to current argument parser.
    `argv`: the command line arguments to parse
    `root_parser`: True if the arguments are assigned to the root parser, else False
    return: a tuple containing [known, rest], where known are the arguments assigned to current parser
    """
    non_help_argv = list(filter(lambda a: a not in DEFAULT_HELP_ARGUMENTS, argv))
    help_argv = list(filter(lambda a: a in DEFAULT_HELP_ARGUMENTS, argv))
    only_help = len(non_help_argv) == 1

    known = non_help_argv + help_argv if only_help else non_help_argv
    rest = [] if only_help else help_argv
    return known, rest


# Cell
def parse_known_args_with_help(parser, argv):
    """
    Parse the command line arguments and validate, if help parameter should be assigned to current argument parser.
    If the help parameter is not assigned to current parser, then keep it in the remaining argv.
    `parser`: the argument parser to handle
    return: a tuple containing [args, argv], where args are the consumed arguments and argv are the remaining arguments
    """
    argv, argv_help_rest = parse_known_help(argv)
    args, rest_args = parser.parse_known_args(argv)
    argv = rest_args + argv_help_rest
    return args, argv


# Cell
def assign_arg_prefix(arg_name, prefix=None):
    """
    Assign a parameter prefix to a given argument name. (e.g --prefix_<arg_name>)
    `arg_name`: the argument name to prefix
    `prefix`: the prefix
    return: the prefixed argument name
    """
    return arg_name if prefix is None else re.sub(r'^(-{0,2})([\w-]+)$', r'\1{}_\2'.format(prefix), arg_name)


# Cell
def input_feedback(msg, choice, choices):
    """
    User input request wrapper.
    `msg`: the message to display
    `choice`: if previous choice exist
    `choices`: the possible choices
    :return: the choice input
    """

    # if decision is already made for all contents, skip feedback
    if not (choice and choice.isupper()):
        prompt = '{} \n choices: {} '.format(msg, ', '.join(['{} ({})'.format(k, v) for k, v in choices.items()]))
        while True:
            choice = input(prompt)
            if choice in choices:
                break
    return choice



# AUTOGENERATED! DO NOT EDIT! File to edit: tensorflow-tflite_converter.ipynb (unless otherwise specified).

__all__ = ['TFLITE_MODEL_DEFAULT_NAME', 'logger', 'convert_model', 'save_model', 'read_pipeline_config',
           'configure_logging']


# Cell
import logging
import logging.handlers
import argparse
import sys
import tensorflow as tf
from os.path import join
from google.protobuf import text_format
from ..dataset.type import DatasetType, infer_dataset_type
from .tflite_metadata import create_metadata, write_metadata


# Cell
TFLITE_MODEL_DEFAULT_NAME = 'model.tflite'


# Cell
logger = logging.getLogger(__name__)


# Cell
def convert_model(saved_model_dir):
    """
    Convert a SavedModel into Tensorflow Lite Format.
    `saved_model_dir`: the path to the SavedModel directory
    returns: the converted Tensorflow Lite model
    """
    logger.info('Converting SavedModel from: {}'.format(saved_model_dir))
    converter = tf.lite.TFLiteConverter.from_saved_model(saved_model_dir)  # path to the SavedModel directory
    tflite_model = converter.convert()
    return tflite_model


# Cell
def save_model(tflite_model, output_file):
    """
    Save a Tensowflow Lite model to disk.
    `tflite_model`: the Tensorflow Lite model
    `output_file`: the path and filename to save the Tensorflow Lite model
    """
    with open(output_file, 'wb') as f:
        f.write(tflite_model)
    logger.info('Successfully save model to file: {}'.format(output_file))


# Cell
def read_pipeline_config(pipeline_config_path):
    """
    Reads the pipeline config file.

    `pipeline_config_path`: The path to the pipeline config file.
    """
    pipeline_config = {}
    with tf.io.gfile.GFile(pipeline_config_path, 'r') as f:
        text_format.Parse(f.read(), pipeline_config)
    return pipeline_config


# Cell
def configure_logging(logging_level=logging.INFO):
    """
    Configures logging for the system.

    `logging_level`: The logging level to use.
    """
    logger.setLevel(logging_level)

    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging_level)

    logger.addHandler(handler)


# Cell
if __name__ == '__main__' and '__file__' in globals():
    configure_logging()

    parser = argparse.ArgumentParser()
    parser.add_argument("-s",
                        "--source",
                        help="The path to the folder containing the SavedModel.",
                        type=str)
    parser.add_argument("-c",
                        "--categories",
                        help="The categories file to add to the Tensorflow Lite model.",
                        type=str)
    parser.add_argument("-n",
                        "--name",
                        help="The name of the model.",
                        type=str)
    parser.add_argument("-v",
                        "--version",
                        help="The version of the model.",
                        type=int,
                        default=1)
    parser.add_argument("-t",
                        "--type",
                        help="The type of the model, if not explicitly set try to infer from categories file path.",
                        choices=list(DatasetType),
                        type=DatasetType,
                        default=None)
    parser.add_argument("-o",
                        "--output",
                        help="The folder to store the Tensorflow Lite model.",
                        type=str)
    args = parser.parse_args()

    model_type = args.type

    # try to infer the model type if not explicitly set
    if model_type is None:
        try:
            model_type = infer_dataset_type(args.categories)
        except ValueError as e:
            logger.error(e)
            sys.exit(1)

    output_file = join(args.output, TFLITE_MODEL_DEFAULT_NAME)

    save_model(convert_model(args.source), output_file)

    model_meta = create_metadata(args.source, args.categories, model_type, args.name, args.version)
    write_metadata(model_meta, output_file, args.categories)

    logger.info('FINISHED!!!')



# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"logger": "tools-image_size_calculator.ipynb",
         "SubsetType": "annotation-core.ipynb",
         "RegionShape": "annotation-core.ipynb",
         "parse_region_shape": "annotation-core.ipynb",
         "Region": "annotation-core.ipynb",
         "Annotation": "annotation-core.ipynb",
         "convert_region": "annotation-core.ipynb",
         "region_bounding_box": "annotation-core.ipynb",
         "AnnotationAdapter": "annotation-core.ipynb",
         "DEFAULT_CATEGORY_FOLDER_INDEX": "annotation-folder_category_adapter.ipynb",
         "FolderCategoryAnnotationAdapter": "annotation-folder_category_adapter.ipynb",
         "DEFAULT_ANNOTATIONS_FILE": "annotation-via_adapter.ipynb",
         "CSV_FIELDNAMES": "annotation-via_adapter.ipynb",
         "MultiCategoryAnnotationAdapter": "annotation-multi_category_adapter.ipynb",
         "DEFAULT_CATEGORY_LABEL_KEY": "annotation-via_adapter.ipynb",
         "VIAAnnotationAdapter": "annotation-via_adapter.ipynb",
         "FIELD_NAMES": "annotation-yolo_adapter.ipynb",
         "DEFAULT_IMAGES_FOLDER": "annotation-yolo_adapter.ipynb",
         "DEFAULT_IMAGE_ANNOTATIONS_FOLDER": "annotation-yolo_adapter.ipynb",
         "YOLOAnnotationAdapter": "annotation-yolo_adapter.ipynb",
         "convert": "annotation_converter.ipynb",
         "configure_logging": "tools-image_size_calculator.ipynb",
         "WINDOW_NAME": "annotation_viewer.ipynb",
         "show_annotated_images": "annotation_viewer.ipynb",
         "DEFAULT_CATEGORIES_FILE": "category_tools.ipynb",
         "NOT_CATEGORIZED": "category_tools.ipynb",
         "BACKGROUND_CLASS": "category_tools.ipynb",
         "BACKGROUND_CLASS_CODE": "category_tools.ipynb",
         "read_categories": "category_tools.ipynb",
         "write_categories": "category_tools.ipynb",
         "DEFAULT_HELP_ARGUMENTS": "core.ipynb",
         "import_modules": "core.ipynb",
         "list_subclasses": "core.ipynb",
         "all_subclasses": "core.ipynb",
         "parse_known_help": "core.ipynb",
         "parse_known_args_with_help": "core.ipynb",
         "assign_arg_prefix": "core.ipynb",
         "input_feedback": "core.ipynb",
         "Dataset": "dataset-core.ipynb",
         "ImageClassificationDataset": "dataset-image_classification.ipynb",
         "ImageObjectDetectionDataset": "dataset-image_object_detection.ipynb",
         "ImageSegmentationDataset": "dataset-image_segmentation.ipynb",
         "DatasetType": "dataset-type.ipynb",
         "infer_dataset_type": "dataset-type.ipynb",
         "DATASET_FOLDER": "dataset_generator.ipynb",
         "generate": "dataset_generator.ipynb",
         "box_area": "evaluation-core.ipynb",
         "intersection_box": "evaluation-core.ipynb",
         "union_box": "evaluation-core.ipynb",
         "intersection_over_union": "evaluation-core.ipynb",
         "create_ellipse": "geometry.ipynb",
         "ellipse_intersection_area": "geometry.ipynb",
         "pascal_voc_palette": "image-color_palette.ipynb",
         "ade20k_palette": "image-color_palette.ipynb",
         "cityscapes_palette": "image-color_palette.ipynb",
         "mhp_palette_v1": "image-color_palette.ipynb",
         "setup_learner": "image-inference.ipynb",
         "fix_odd_sides": "image-inference.ipynb",
         "Inference": "image-inference.ipynb",
         "ClassificationInference": "image-inference.ipynb",
         "SegmentationInference": "image-inference.ipynb",
         "limit_to_max_size": "image-pillow_tools.ipynb",
         "fit_to_max_size": "image-pillow_tools.ipynb",
         "get_image_size": "image-pillow_tools.ipynb",
         "EXIF_ORIENTATION_TAG": "image-pillow_tools.ipynb",
         "get_image_orientation": "image-pillow_tools.ipynb",
         "read_exif_metadata": "image-pillow_tools.ipynb",
         "write_exif_metadata": "image-pillow_tools.ipynb",
         "assign_exif_orientation": "image-pillow_tools.ipynb",
         "convert_to_base64": "image-pillow_tools.ipynb",
         "write_mask": "image-pillow_tools.ipynb",
         "ImageOrientation": "image-tools.ipynb",
         "get_image_scale": "image-tools.ipynb",
         "BUF_SIZE": "tools-check_double_images.ipynb",
         "DEFAULT_IMAGE_EXTENSIONS": "io-core.ipynb",
         "FOLDER_FILTER": "io-core.ipynb",
         "scan_files": "io-core.ipynb",
         "scan_path": "io-core.ipynb",
         "delete_folder": "io-core.ipynb",
         "create_folder": "io-core.ipynb",
         "get_file_sha": "tools-check_double_images.ipynb",
         "assign_filename_suffix": "io-core.ipynb",
         "strip_path": "io-core.ipynb",
         "TFLITE_MODEL_DEFAULT_NAME": "tensorflow-tflite_converter.ipynb",
         "convert_model": "tensorflow-tflite_converter.ipynb",
         "save_model": "tensorflow-tflite_converter.ipynb",
         "read_pipeline_config": "tensorflow-tflite_converter.ipynb",
         "SAVED_MODEL_META_DEFAULT_KEY": "tensorflow-tflite_metadata.ipynb",
         "AUTHOR": "tensorflow-tflite_metadata.ipynb",
         "MetaInfo": "tensorflow-tflite_metadata.ipynb",
         "create_metadata": "tensorflow-tflite_metadata.ipynb",
         "write_metadata": "tensorflow-tflite_metadata.ipynb",
         "read_metadata": "tensorflow-tflite_metadata.ipynb",
         "create_tfrecord_entry": "tensorflow-tfrecord_builder.ipynb",
         "create_tfrecord_file": "tensorflow-tfrecord_builder.ipynb",
         "int64_feature": "tensorflow-tfrecord_builder.ipynb",
         "int64_list_feature": "tensorflow-tfrecord_builder.ipynb",
         "bytes_feature": "tensorflow-tfrecord_builder.ipynb",
         "bytes_list_feature": "tensorflow-tfrecord_builder.ipynb",
         "float_list_feature": "tensorflow-tfrecord_builder.ipynb",
         "create_labelmap_file": "tensorflow-tfrecord_builder.ipynb",
         "FILE_FILTER": "tools-check_double_images.ipynb",
         "IMAGE_EXTENSION": "tools-check_double_images.ipynb",
         "DoubleFileChecker": "tools-check_double_images.ipynb",
         "check_double_entries": "tools-check_double_images.ipynb",
         "remove_entries": "tools-check_double_images.ipynb",
         "scan_folder": "tools-check_double_images.ipynb",
         "Downloader": "tools-downloader.ipynb",
         "SaveToDirectory": "tools-downloader.ipynb",
         "ImageSizeCalculator": "tools-image_size_calculator.ipynb"}

modules = ["annotation/core.py",
           "annotation/folder_category_adapter.py",
           "annotation/multi_category_adapter.py",
           "annotation/via_adapter.py",
           "annotation/yolo_adapter.py",
           "annotation_converter.py",
           "annotation_viewer.py",
           "category_tools.py",
           "core.py",
           "dataset/core.py",
           "dataset/image_classification.py",
           "dataset/image_object_detection.py",
           "dataset/image_segmentation.py",
           "dataset/type.py",
           "dataset_generator.py",
           "evaluation/core.py",
           "geometry.py",
           "image/color_palette.py",
           "image/inference.py",
           "image/opencv_tools.py",
           "image/pillow_tools.py",
           "image/tools.py",
           "io/core.py",
           "tensorflow/tflite_converter.py",
           "tensorflow/tflite_metadata.py",
           "tensorflow/tfrecord_builder.py",
           "tools/check_double_images.py",
           "tools/downloader.py",
           "tools/image_size_calculator.py"]

doc_url = "https://AI-Force.github.io/aiforce/"

git_url = "https://github.com/AI-Force/aiforce/tree/master/"


def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: image-tools.ipynb (unless otherwise specified).

__all__ = ['ImageOrientation', 'get_image_scale']


# Cell
from enum import Enum


# Cell
class ImageOrientation(Enum):
    """
    Image EXIF orientations.
    """
    TOP = 1
    TOP_FLIPPED = 2
    BOTTOM = 3
    BOTTOM_FLIPPED = 4
    RIGHT_FLIPPED = 5
    RIGHT = 6
    LEFT_FLIPPED = 7
    LEFT = 8

    def __str__(self):
        return str(self.value)


# Cell
def get_image_scale(image_size, target_size):
    """
    Calculates the scale of the image to fit the target size.
    `image_size`: The image size as tuple of (w, h)
    `target_size`: The target size as tuple of (w, h).
    :return: The image scale as tuple of (w_scale, h_scale)
    """
    (image_w, image_h) = image_size
    (target_w, target_h) = target_size
    scale = (target_w / float(image_w), target_h / float(image_h))
    return scale



# AUTOGENERATED! DO NOT EDIT! File to edit: tools-downloader.ipynb (unless otherwise specified).

__all__ = ['Downloader', 'SaveToDirectory']


# Cell
import errno
import logging
import binascii
import os
import os.path
import io
import requests
from multiprocessing.dummy import Pool as ThreadPool


# Cell
class Downloader:
    """Multi-threaded downloader that can retry a download on failure."""

    _DEFAULT_THREAD_COUNT = 10
    """The default number of threads to use when downloading images."""
    _DEFAULT_RETRIES = 3
    """The number of times to attempt to download the file before giving up."""
    _DEFAULT_TIMEOUT = 30
    """The number of seconds to wait when attempting to create a connection."""

    _SERVER_ERROR_CODE = 500
    """Represents a server error."""
    _FILE_NOT_FOUND = 404
    """Represents that a file could not be found on the remote server."""

    _RANDOM_FILE_NAME_LENGTH = 15
    """Length of the randomly generated filename."""

    def __init__(self, retries=_DEFAULT_RETRIES, thread_count=_DEFAULT_THREAD_COUNT,
                 timeout=_DEFAULT_TIMEOUT):
        """Creates a new Downloaded object with an empty list of downloads."""
        self.tasks = []
        self.retries = retries
        self.thread_count = thread_count
        self.output_paths = []
        self.timeout = timeout
        self.failed_downloads = []
        self.callbacks = []

    def add_callback(self, callback):
        """
        Adds a callback that will be run after the download of the file.

        :param callback: The callback to add.
        """
        self.callbacks.append(callback)

    def _random_file_name(self):
        """Generates a random file name. This is used in cases where the file
        cannot be determined from the URL alone."""
        return binascii.b2a_hex(os.urandom(self._RANDOM_FILE_NAME_LENGTH))

    def _guess_file_name(self, url):
        """Guesses the file name by first attempting to extract the last part
        of the URL, after the slash."""
        file_name = url.rpartition('/')[-1]
        if not file_name:
            file_name = self._random_file_name()
        return file_name

    @staticmethod
    def _get_file_extension(url_part):
        """Returns the file extension of the specified URL."""
        return os.path.splitext(url_part)[-1]

    def add_task(self, url, index):
        """Adds a task to the list of tasks to download."""
        logging.debug("Added Task, URL: %s, INDEX: %d.", (url, index))
        self.tasks.append((url, index))

    def add_tasks(self, urls, indexes):
        """Adds a list of tasks to the list of tasks to download."""
        self.tasks.extend(list(zip(urls, indexes)))

    def _do_download(self, task):
        """Performs the actual download of the file."""
        retry_count = 0
        downloaded = False
        url, index = task
        while not downloaded and retry_count < self.retries:
            try:
                logging.info("Downloading %s.", url)
                r = requests.get(url, stream=True, timeout=self.timeout)
                r.raise_for_status()
                raw_data = io.BytesIO(r.content)
                for callback in self.callbacks:
                    callback.perform(index, raw_data)
                downloaded = True
            except requests.exceptions.HTTPError as e:
                logging.error("Communication occurred during download.")
                logging.exception(e)
                if e.response.status_code == self._FILE_NOT_FOUND:
                    logging.error("Remote file does not exist; skipping.")
                    break
                if e.response.status_code == self._SERVER_ERROR_CODE:
                    retry_count += 1
                    logging.warning("Retry attempt %d", retry_count)
                    continue
                else:
                    logging.error("Unhandled HTTP error occurred during download.")
                    logging.exception(e)
                    break
            except Exception as e:
                logging.error("Unknown error occurred during download.")
                logging.exception(e)
                break
        if not downloaded:
            self.failed_downloads.append(task)

    def download(self):
        """Downloads the files to the specified location."""
        pool = ThreadPool(self.thread_count)
        pool.map(self._do_download, self.tasks)
        pool.close()
        pool.join()

    def clear(self):
        """Removes all of the downloads from the list of downloads."""
        del self.tasks[:]
        del self.failed_downloads[:]

    def get_failed_downloads(self):
        """Returns a list of downloads that failed to complete."""
        return self.failed_downloads


# Cell
class SaveToDirectory:
    """Callback to download file to a specific directory."""
    def __init__(self, file_path):
        """
        Initializes the object to save to a specific directory.

        :param file_path:  The base path to save the file to.
        """
        self.path = file_path
        self._make_path()

    def _make_path(self):
        """
        Ensures a destination path exists by creating it.
        """
        try:
            os.makedirs(self.path)
        except OSError as e:
            if e.errno == errno.EEXIST and os.path.isdir(self.path):
                pass
            else:
                raise

    def perform(self, filename, data):
        """
        Saves the data to the specified file.

        :param filename: The filename to write the file to.
        :param data: The data to save as a BytesIO object.
        """
        save_path = os.path.join(self.path, filename)
        with open(save_path, "wb") as f:
            f.write(data.getvalue())



# AUTOGENERATED! DO NOT EDIT! File to edit: annotation_viewer.ipynb (unless otherwise specified).

__all__ = ['WINDOW_NAME', 'show_annotated_images', 'configure_logging']


# Cell
import sys
import argparse
import logging
import cv2
import numpy as np
from .core import list_subclasses, parse_known_args_with_help
from aiforce import annotation as annotation_package
from .annotation.core import AnnotationAdapter, SubsetType
from .image.opencv_tools import fit_to_max_size
from .annotation.core import RegionShape


# Cell
WINDOW_NAME = 'Annotation'


# Cell
def show_annotated_images(annotation_adapter, max_width=0, max_height=0):
    """
    Show images with corresponding annotations.
    Images are shown one at a time with switching by using the arrow left/right keys.
    `annotation_adapter`: The annotation adapter to use
    `max_width`: The maximum width to scale the image for visibility.
    `max_height`: The maximum height to scale the image for visibility.
    """
    categories = annotation_adapter.read_categories()
    annotations = annotation_adapter.read_annotations(SubsetType.TRAINVAL)
    len_annotations = len(annotations)

    if len_annotations == 0:
        logging.error("No Annotations found")
        return

    index = 0
    annotation_keys = list(annotations.keys())

    logging.info("Keys to use:")
    logging.info("n = Next Image")
    logging.info("b = Previous Image")
    logging.info("q = Quit")

    logging.info("Annotations to view: {}".format(len_annotations))

    while True:
        annotation_id = annotation_keys[index]
        annotation = annotations[annotation_id]
        logging.info("View Image {}/{}: {}".format(index + 1, len_annotations, annotation.file_path))
        img = cv2.imread(annotation.file_path, cv2.IMREAD_COLOR)

        if img is None:
            logging.info("Image not found at {}".format(annotation.file_path))
            img = np.zeros(shape=(1, 1, 3))

        if annotation.regions:
            logging.info("Found {} regions".format(len(annotation.regions)))
            for region_index, region in enumerate(annotation.regions):
                points = list(zip(region.points_x, region.points_y))
                logging.info("Found {} of category {} with {} points: {}".format(region.shape,
                                                                                 ','.join(region.labels),
                                                                                 len(points), points))
                if region.shape == RegionShape.CIRCLE:
                    img = cv2.circle(img, points[0], region.radius_x, (0, 255, 255), 2)
                elif region.shape == RegionShape.ELLIPSE:
                    img = cv2.ellipse(img, points[0], (region.radius_x, region.radius_y), 0, 0, 360,
                                      (0, 255, 255), 2)
                elif region.shape == RegionShape.POINT:
                    img = cv2.circle(img, points[0], 1, (0, 255, 255), 2)
                elif region.shape == RegionShape.POLYGON:
                    pts = np.array(points, np.int32)
                    pts = pts.reshape((-1, 1, 2))
                    img = cv2.polylines(img, [pts], True, (0, 255, 255), 2)
                elif region.shape == RegionShape.RECTANGLE:
                    img = cv2.rectangle(img, points[0], points[1], (0, 255, 255), 2)

        if max_width and max_height:
            img = fit_to_max_size(img, max_width, max_height)

        cv2.imshow(WINDOW_NAME, img)
        cv2.setWindowTitle(WINDOW_NAME, "Image {}/{}".format(index + 1, len_annotations))

        k = cv2.waitKey(0)
        if k == ord('q'):    # 'q' key to stop
            break
        elif k == ord('b'):
            index = max(0, index - 1)
        elif k == ord('n'):
            index = min(len_annotations - 1, index + 1)

    cv2.destroyAllWindows()


# Cell
def configure_logging(logging_level=logging.INFO):
    """
    Configures logging for the system.

    :param logging_level: The logging level to use.
    """
    logging.basicConfig(level=logging_level)


# Cell
if __name__ == '__main__' and '__file__' in globals():
    # for direct shell execution
    configure_logging()

    # read annotation adapters to use
    adapters = list_subclasses(annotation_package, AnnotationAdapter)

    parser = argparse.ArgumentParser()
    parser.add_argument("-a",
                        "--annotation",
                        help="The annotation adapter to read the annotations.",
                        type=str,
                        choices=adapters.keys(),
                        required=True)
    parser.add_argument("--max-width",
                        help="The maximum width to scale the image for visibility.",
                        type=int,
                        default=0)
    parser.add_argument("--max-height",
                        help="The maximum height to scale the image for visibility.",
                        type=int,
                        default=0)

    argv = sys.argv
    args, argv = parse_known_args_with_help(parser, argv)

    adapter_class = adapters[args.annotation]

    # parse the annotation arguments
    annotation_parser = getattr(adapter_class, 'argparse')()
    annotation_args, argv = parse_known_args_with_help(annotation_parser, argv)

    show_annotated_images(adapter_class(**vars(annotation_args)), args.max_width, args.max_height)



# AUTOGENERATED! DO NOT EDIT! File to edit: category_tools.ipynb (unless otherwise specified).

__all__ = ['DEFAULT_CATEGORIES_FILE', 'NOT_CATEGORIZED', 'BACKGROUND_CLASS', 'BACKGROUND_CLASS_CODE', 'logger',
           'read_categories', 'write_categories', 'configure_logging']


# Cell
import argparse
import logging
from os.path import isfile
from .dataset.type import DatasetType


# Cell
DEFAULT_CATEGORIES_FILE = 'categories.txt'
NOT_CATEGORIZED = '[NOT_CATEGORIZED]'
BACKGROUND_CLASS = '_background_'
BACKGROUND_CLASS_CODE = 0


# Cell

logger = logging.getLogger(__name__)


# Cell
def read_categories(categories_file=None, dataset_type=DatasetType.IMAGE_CLASSIFICATION):
    """
    Reads the categories from a categories file.
    If the dataset type is image segmentation or object detection, a background class at index 0 is prepend.
    If the optional `categories_file` is not given, the file name *categories.txt* is used by default
    `categories_file`: the categories file name, if not the default
    `dataset_type`: the type of the data-set to create the categories for
    return: a list of the category names
    """
    if categories_file is None:
        categories_file = DEFAULT_CATEGORIES_FILE

    if not isfile(categories_file):
        logger.warning('Categories file not found at: {}'.format(categories_file))
        return []
    with open(categories_file) as f:
        categories = f.read().strip().split('\n')
        logger.info('Read {} categories from categories file at: {}'.format(len(categories), categories_file))
    if dataset_type in [DatasetType.IMAGE_OBJECT_DETECTION, DatasetType.IMAGE_SEGMENTATION]:
        categories = [BACKGROUND_CLASS] + categories
        logger.info('Prepend background class {} to the categories'.format(BACKGROUND_CLASS))

    return categories


# Cell
def write_categories(categories, categories_file=None):
    """
    Write the categories to a categories file.
    If the dataset type is image segmentation or object detection, a background class at index 0 is prepend.
    If the optional `categories_file` is not given, the file name *categories.txt* is used by default
    `categories`: a list of the category names to write
    `categories_file`: the categories file name
    """
    if categories_file is None:
        categories_file = DEFAULT_CATEGORIES_FILE

    if len(categories) > BACKGROUND_CLASS_CODE and categories[BACKGROUND_CLASS_CODE] == BACKGROUND_CLASS:
        logger.info('Remove background class {} from the categories'.format(BACKGROUND_CLASS))
        categories = categories[1:]
    with open(categories_file, 'w') as f:
        f.write('\n'.join(categories))
        logger.info('Write {} categories to categories file at: {}'.format(len(categories), categories_file))



# Cell
def configure_logging(logging_level=logging.INFO):
    """
    Configures logging for the system.

    :param logging_level: The logging level to use.
    """
    logging.basicConfig(level=logging_level)


# Cell
if __name__ == '__main__' and '__file__' in globals():
    # for direct shell execution
    configure_logging()

    parser = argparse.ArgumentParser()
    parser.add_argument("categories",
                        help="The path to the categories file.")

    args = parser.parse_args()

    print(read_categories(args.categories))



# AUTOGENERATED! DO NOT EDIT! File to edit: annotation-yolo_adapter.ipynb (unless otherwise specified).

__all__ = ['FIELD_NAMES', 'DEFAULT_IMAGES_FOLDER', 'DEFAULT_IMAGE_ANNOTATIONS_FOLDER', 'logger',
           'YOLOAnnotationAdapter']


# Cell
import csv
import logging
import shutil
from os.path import join, splitext, basename, isfile
from ..category_tools import read_categories
from ..core import assign_arg_prefix
from ..io.core import scan_files, create_folder
from ..image.pillow_tools import get_image_size
from .core import Annotation, AnnotationAdapter, Region, RegionShape, SubsetType


# Cell
FIELD_NAMES = ['class_number', 'c_x', 'c_y', 'width', 'height']
DEFAULT_IMAGES_FOLDER = 'images'
DEFAULT_IMAGE_ANNOTATIONS_FOLDER = 'labels'


# Cell
logger = logging.getLogger(__name__)


# Cell
class YOLOAnnotationAdapter(AnnotationAdapter):
    """
    Adapter to read and write annotations in the YOLO format.
    """

    def __init__(self, path, categories_file_name=None, images_folder_name=None, annotations_folder_name=None):
        """
        YOLO Adapter to read and write annotations.
        `path`: the folder containing the annotations
        `categories_file_name`: the name of the categories file
        `images_folder_name`: the name of the folder containing the image files
        `annotations_folder_name`: the name of the folder containing the mage annotations
        """
        super().__init__(path, categories_file_name)

        if images_folder_name is None:
            self.images_folder_name = DEFAULT_IMAGES_FOLDER
        else:
            self.images_folder_name = images_folder_name

        if annotations_folder_name is None:
            self.annotations_folder_name = DEFAULT_IMAGE_ANNOTATIONS_FOLDER
        else:
            self.annotations_folder_name = annotations_folder_name

    @classmethod
    def argparse(cls, prefix=None):
        """
        Returns the argument parser containing argument definition for command line use.
        `prefix`: a parameter prefix to set, if needed
        return: the argument parser
        """
        parser = super(YOLOAnnotationAdapter, cls).argparse(prefix=prefix)
        parser.add_argument(assign_arg_prefix('--images_folder_name', prefix),
                            dest="images_folder_name",
                            help="The name of the folder containing the image files.",
                            default=None)
        parser.add_argument(assign_arg_prefix('--annotations_folder_name', prefix),
                            dest="annotations_folder_name",
                            help="The name of the folder containing the mage annotations.",
                            default=None)
        return parser

    def read_annotations(self, subset_type=SubsetType.TRAINVAL):
        """
        Reads YOLO annotations.
        `subset_type`: the subset type to read
        return: the annotations as dictionary
        """
        path = join(self.path, str(subset_type))
        annotations = {}
        annotations_path = join(path, self.annotations_folder_name)
        images_path = join(path, self.images_folder_name)
        logger.info('Read images from {}'.format(images_path))
        logger.info('Read annotations from {}'.format(annotations_path))

        annotation_files = scan_files(annotations_path, file_extensions='.txt')
        categories = read_categories(join(self.path, self.categories_file_name))
        categories_len = len(categories)
        skipped_annotations = []

        for annotation_file in annotation_files:
            with open(annotation_file, newline='') as csv_file:
                annotation_file_name = basename(annotation_file)
                file_name, _ = splitext(annotation_file_name)
                file_path = join(images_path, '{}{}'.format(file_name, '.jpg'))

                if not isfile(file_path):
                    logger.warning("{}: Source file not found, skip annotation.".format(file_path))
                    skipped_annotations.append(file_path)
                    continue

                if annotation_file not in annotations:
                    annotations[annotation_file] = Annotation(annotation_id=annotation_file, file_path=file_path)

                annotation = annotations[annotation_file]

                reader = csv.DictReader(csv_file, fieldnames=FIELD_NAMES, delimiter=' ')
                _, image_width, image_height = get_image_size(file_path)
                for row in reader:
                    c_x = float(row["c_x"])
                    c_y = float(row["c_y"])
                    width = float(row["width"])
                    height = float(row["height"])
                    class_number = int(row["class_number"])
                    # denormalize bounding box
                    x_min = self._denormalize_value(c_x - (width / 2), image_width)
                    y_min = self._denormalize_value(c_y - (height / 2), image_height)
                    x_max = x_min + self._denormalize_value(width, image_width)
                    y_max = y_min + self._denormalize_value(height, image_height)
                    points_x = [x_min, x_max]
                    points_y = [y_min, y_max]

                    labels = [categories[class_number]] if class_number < categories_len else []
                    if not labels:
                        logger.warning("{}: Class number exceeds categories, set label as empty.".format(
                            annotation_file
                        ))
                    region = Region(shape=RegionShape.RECTANGLE, points_x=points_x, points_y=points_y, labels=labels)
                    annotation.regions.append(region)

        logger.info('Finished read annotations')
        logger.info('Annotations read: {}'.format(len(annotations)))
        if skipped_annotations:
            logger.info('Annotations skipped: {}'.format(len(skipped_annotations)))

        return annotations

    def write_annotations(self, annotations, subset_type=SubsetType.TRAINVAL):
        """
        Writes YOLO annotations to the annotations folder and copy the corresponding source files.
        `annotations`: the annotations as dictionary
        `subset_type`: the subset type to write
        return: a list of written target file paths
        """
        path = join(self.path, str(subset_type))
        create_folder(path)
        annotations_path = join(path, self.annotations_folder_name)
        annotations_folder = create_folder(annotations_path)
        images_path = join(path, self.images_folder_name)
        images_folder = create_folder(images_path)
        categories = read_categories(join(self.path, self.categories_file_name))

        logger.info('Write images to {}'.format(images_folder))
        logger.info('Write annotations to {}'.format(annotations_folder))

        copied_files = []
        skipped_annotations = []

        for annotation in annotations.values():
            annotation_file_name = basename(annotation.file_path)
            file_name, _ = splitext(annotation_file_name)
            annotations_file = join(annotations_folder, '{}{}'.format(file_name, '.txt'))
            target_file = join(images_folder, annotation_file_name)

            if not isfile(annotation.file_path):
                logger.warning("{}: Source file not found, skip annotation.".format(annotation.file_path))
                skipped_annotations.append(annotation.file_path)
                continue
            if isfile(target_file):
                logger.warning("{}: Target file already exist, skip annotation.".format(annotation.file_path))
                skipped_annotations.append(annotation.file_path)
                continue

            _, image_width, image_height = get_image_size(annotation.file_path)
            rows = []
            skipped_regions = []
            for index, region in enumerate(annotation.regions):
                if region.shape != RegionShape.RECTANGLE:
                    logger.warning('Unsupported shape {}, skip region {} at path: {}'.format(region.shape,
                                                                                             index,
                                                                                             annotations_file))
                    skipped_regions.append(region)
                    continue

                x_min, x_max = region.points_x
                y_min, y_max = region.points_y
                width = x_max - x_min
                height = y_max - y_min
                # normalize bounding box
                c_x = self._normalize_value(x_min + width / 2, image_width)
                c_y = self._normalize_value(y_min + height / 2, image_height)
                width = self._normalize_value(width, image_width)
                height = self._normalize_value(height, image_height)
                label = region.labels[0] if len(region.labels) else ''
                try:
                    class_number = categories.index(label)
                except ValueError:
                    logger.warning('Unsupported label {}, skip region {} at path: {}'.format(label,
                                                                                             index,
                                                                                             annotations_file))
                    skipped_regions.append(region)
                    continue
                rows.append(dict(zip(FIELD_NAMES, [class_number, c_x, c_y, width, height])))

            if len(skipped_regions) == len(annotation.regions):
                logger.warning("{}: All regions skipped, skip annotation.".format(annotation.file_path))
                skipped_annotations.append(annotation.file_path)
                continue

            # write the annotations
            with open(annotations_file, 'w', newline='') as csv_file:
                writer = csv.DictWriter(csv_file, fieldnames=FIELD_NAMES, delimiter=' ')
                writer.writerows(rows)

            # copy the file
            shutil.copy2(annotation.file_path, target_file)
            copied_files.append(target_file)

        logger.info('Finished write annotations')
        logger.info('Annotations written: {}'.format(len(annotations) - len(skipped_annotations)))
        if skipped_annotations:
            logger.info('Annotations skipped: {}'.format(len(skipped_annotations)))
        return copied_files

    @classmethod
    def _denormalize_value(cls, value, metric):
        """
        Denormalize a bounding box value
        `value`: the value to denormalize
        `metric`: the metric to denormalize from
        return: the denormalized value
        """
        return int(value * metric)

    @classmethod
    def _normalize_value(cls, value, metric):
        """
        Normalize a bounding box value
        `value`: the value to normalize
        `metric`: the metric to normalize against
        return: the normalized value
        """
        return float(value) / metric



# AUTOGENERATED! DO NOT EDIT! File to edit: dataset_generator.ipynb (unless otherwise specified).

__all__ = ['DATASET_FOLDER', 'logger', 'configure_logging', 'generate']


# Cell
import sys
import argparse
import logging
from os.path import join
from datetime import datetime
from logging.handlers import MemoryHandler
from .core import list_subclasses, parse_known_args_with_help
from aiforce import annotation as annotation_package
from .annotation.core import AnnotationAdapter
from aiforce import dataset as dataset_package
from .dataset.core import Dataset


# Cell
DATASET_FOLDER = 'datasets'


# Cell
logger = logging.getLogger('mlcore.dataset')


# Cell
def configure_logging(logging_level=logging.INFO):
    """
    Configures logging for the system.
    """
    logging.basicConfig(level=logging_level)

    log_memory_handler = MemoryHandler(1, flushLevel=logging_level)
    log_memory_handler.setLevel(logging_level)

    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setLevel(logging_level)

    logger.addHandler(log_memory_handler)
    logger.addHandler(stdout_handler)

    logger.setLevel(logging_level)

    return log_memory_handler


# Cell
def generate(dataset: Dataset, log_memory_handler):
    """
    Generate a dataset.
    `dataset`: the dataset to build
    `log_memory_handler`: the log handler for the build log
    """
    dataset.build_info()

    logger.info('Start build {} at {}'.format(type(dataset).__name__, dataset.output_adapter.path))

    dataset.create_folders()

    # create the build log file
    log_file_name = datetime.now().strftime("build_%Y.%m.%d-%H.%M.%S.log")
    file_handler = logging.FileHandler(join(dataset.output_adapter.path, log_file_name), encoding="utf-8")
    log_memory_handler.setTarget(file_handler)

    dataset.build()

    logger.info('Finished build {} at {}'.format(type(dataset).__name__, dataset.output_adapter.path))


# Cell
if __name__ == '__main__' and '__file__' in globals():
    # for direct shell execution
    log_handler = configure_logging()

    # read annotation adapters to use
    adapters = list_subclasses(annotation_package, AnnotationAdapter)

    # read datasets to use
    datasets = list_subclasses(dataset_package, Dataset)

    parser = argparse.ArgumentParser()
    parser.add_argument("-i",
                        "--input",
                        help="The annotation input adapter.",
                        type=str,
                        choices=adapters.keys(),
                        required=True)
    parser.add_argument("-d",
                        "--dataset",
                        help="The dataset to generate.",
                        type=str,
                        choices=datasets.keys(),
                        required=True)
    parser.add_argument("-o",
                        "--output",
                        help="The annotation output adapter.",
                        type=str,
                        choices=adapters.keys(),
                        required=True)

    argv = sys.argv
    args, argv = parse_known_args_with_help(parser, argv)
    input_adapter_class = adapters[args.input]
    dataset_class = datasets[args.dataset]
    output_adapter_class = adapters[args.output]

    # parse the input arguments
    input_parser = getattr(input_adapter_class, 'argparse')(prefix='input')
    input_args, argv = parse_known_args_with_help(input_parser, argv)

    # parse the dataset arguments
    dataset_parser = getattr(dataset_class, 'argparse')()
    dataset_args, argv = parse_known_args_with_help(dataset_parser, argv)

    # parse the output arguments
    output_parser = getattr(output_adapter_class, 'argparse')(prefix='output')
    output_args, argv = parse_known_args_with_help(output_parser, argv)

    input_adapter = input_adapter_class(**vars(input_args))
    output_adapter = output_adapter_class(**vars(output_args))
    dataset_args.input_adapter = input_adapter
    dataset_args.output_adapter = output_adapter
    target_dataset = dataset_class(**vars(dataset_args))

    logger.info('Build parameters:')
    logger.info(' '.join(sys.argv[1:]))

    generate(target_dataset, log_handler)



# AUTOGENERATED! DO NOT EDIT! File to edit: dataset-image_classification.ipynb (unless otherwise specified).

__all__ = ['logger', 'ImageClassificationDataset']


# Cell

import logging
from os.path import join
from ..core import assign_arg_prefix
from ..annotation.core import AnnotationAdapter
from .core import Dataset
from ..tensorflow.tfrecord_builder import create_labelmap_file


# Cell

logger = logging.getLogger(__name__)


# Cell


class ImageClassificationDataset(Dataset):
    """
    Classification dataset.
    """

    def __init__(self, input_adapter: AnnotationAdapter, output_adapter: AnnotationAdapter, split=None, seed=None,
                 sample=None, tfrecord=False):
        super().__init__(input_adapter, output_adapter, split, seed, sample)
        self.tfrecord = tfrecord

    @classmethod
    def argparse(cls, prefix=None):
        """
        Returns the argument parser containing argument definition for command line use.
        `prefix`: a parameter prefix to set, if needed
        return: the argument parser
        """
        parser = super(ImageClassificationDataset, cls).argparse(prefix=prefix)
        parser.add_argument(assign_arg_prefix("--tfrecord", prefix),
                            dest="tfrecord",
                            help="Also create .tfrecord files.",
                            action="store_true")
        return parser

    def copy(self, train_annotation_keys, val_annotation_keys, test_files=None):
        """
        Copy the images to the dataset and remove EXIF orientation information by hard-rotate the images.
        If tfrecords are build, generate a labelmap.pbtxt file.
        `train_annotation_keys`: The list of training annotation keys
        `val_annotation_keys`: The list of validation annotation keys
        `test_files`: The list of test file paths
        return: A tuple containing train, val and test target file paths
        """

        train_targets, val_targets, test_targets = super().copy(train_annotation_keys, val_annotation_keys, test_files)

        files = train_targets + val_targets + test_targets
        logger.info('Start assign image orientation to {} images'.format(len(files)))
        for file in files:
            self.assign_orientation(file)
        logger.info('Finished assign image orientation to {} images'.format(len(files)))

        # if create tfrecord, create a labelmap.pbtxt file containing the categories
        if self.tfrecord:
            labelmap_file_name = 'label_map.pbtxt'
            labelmap_output_file = join(self.output_adapter.path, labelmap_file_name)
            logger.info('Generate {}'.format(labelmap_output_file))
            create_labelmap_file(labelmap_output_file, list(self.categories), 1)

        return train_targets, val_targets, test_targets

    def build_info(self):
        """
        Converts annotations
        """
        super().build_info()
        logger.info('create_tfrecord: {}'.format(self.tfrecord))



# AUTOGENERATED! DO NOT EDIT! File to edit: dataset-image_segmentation.ipynb (unless otherwise specified).

__all__ = ['logger', 'ImageSegmentationDataset']


# Cell
import numpy as np
import logging
from os.path import join, isfile, splitext, basename
from functools import partial
from ..core import assign_arg_prefix, input_feedback
from ..annotation.core import AnnotationAdapter
from .image_object_detection import ImageObjectDetectionDataset
from ..image import color_palette
from ..image.pillow_tools import assign_exif_orientation, get_image_size, write_mask
from ..io.core import create_folder
from ..annotation.core import RegionShape, convert_region, region_bounding_box


# Cell
logger = logging.getLogger(__name__)


# Cell
class ImageSegmentationDataset(ImageObjectDetectionDataset):

    SEMANTIC_MASK_FOLDER = 'semantic_masks'

    def __init__(self, input_adapter: AnnotationAdapter, output_adapter: AnnotationAdapter, split=None, seed=None,
                 sample=None, tfrecord=False, join_overlapping_regions=False, annotation_area_threshold=None,
                 generate_semantic_masks=True, palette=color_palette.__all__[0]):
        super().__init__(input_adapter, output_adapter, split, seed, sample, tfrecord, join_overlapping_regions,
                         annotation_area_threshold)
        self.generate_semantic_masks = generate_semantic_masks
        self.palette = getattr(color_palette, palette)()
        self.semantic_mask_folder = join(self.output_adapter.path, self.SEMANTIC_MASK_FOLDER)

    @classmethod
    def argparse(cls, prefix=None):
        """
        Returns the argument parser containing argument definition for command line use.
        `prefix`: a parameter prefix to set, if needed
        return: the argument parser
        """
        parser = super(ImageSegmentationDataset, cls).argparse(prefix=prefix)
        parser.add_argument(assign_arg_prefix("--generate_semantic_masks", prefix),
                            dest="generate_semantic_masks",
                            help="Whether semantic masks should be generated.",
                            action="store_true",
                            default=True)
        parser.add_argument("--palette",
                            help="The color palette to use.",
                            choices=color_palette.__all__,
                            default=color_palette.__all__[0])
        return parser

    def create_folders(self):
        """
        Creates the data-set folder structure, if not exist
        """
        super().create_folders()

        if self.generate_semantic_masks:
            # create semantic mask file folder and remove previous data if exist
            semantic_mask_folder = create_folder(self.semantic_mask_folder, clear=True)
            logger.info("Created semantic mask folder {}".format(semantic_mask_folder))

    def copy(self, train_annotation_keys, val_annotation_keys, test_files=None):
        """
        Copy the images to the dataset and remove EXIF orientation information by hard-rotate the images.
        If tfrecords should be build, create tfrecords for train and val subsets and generate a labelmap.pbtxt file.
        If semantic masks should be generate, masks for train and val subsets are build.
        `train_annotation_keys`: The list of training annotation keys
        `val_annotation_keys`: The list of validation annotation keys
        `test_files`: The list of test file paths
        return: A tuple containing train, val and test target file paths
        """

        train_targets, val_targets, test_targets = super().copy(train_annotation_keys, val_annotation_keys, test_files)

        if self.generate_semantic_masks:
            # save semantic masks
            self._save_semantic_masks(train_annotation_keys + val_annotation_keys)

        return train_targets, val_targets, test_targets

    def convert_annotations(self):
        """
        Converts segmentation regions from rectangle to polygon, if exist
        """

        # only the trainval images have annotation, not the test images
        steps = [
            {
                'name': 'position',
                'choices': {
                    's': 'Skip',  # just delete the annotation
                    'S': 'Skip All',
                    't': 'Trim',  # transform the annotation
                    'T': 'Trim All',
                },
                'choice': None,
                'condition': lambda p_min, p_max, size: p_min < 0 or p_max >= size,
                'message': '{} -> {} : {}Exceeds image {}. \n Points \n x: {} \n y: {}',
                'transform': lambda p, size=0: max(min(p, size - 1), 0),
            },
            {
                'name': 'size',
                'choices': {
                    's': 'Skip',  # just delete the annotation
                    'S': 'Skip All',
                    'k': 'Keep',  # transform the annotation (in this case do nothing)
                    'K': 'Keep All',
                },
                'choice': None,
                'condition': lambda p_min, p_max, _: p_max - p_min <= 1,
                'message': '{} -> {} : {}Shape {} is <= 1 pixel. \n Points \n x: {} \n y: {}',
                'transform': lambda p, size=0: p,
            }
        ]

        logger.info('Start convert image annotations from {}'.format(self.input_adapter.path))

        for annotation in self.annotations.values():
            # skip file, if regions are empty or file do not exist
            if not (annotation.regions and isfile(annotation.file_path)):
                continue

            image, _, __ = assign_exif_orientation(annotation.file_path)
            image_width, image_height = image.size

            delete_regions = {}
            for index, region in enumerate(annotation.regions):
                for step in steps:
                    # validate the shape size
                    (x_min, x_max), (y_min, y_max) = region_bounding_box(region)

                    width_condition = step['condition'](x_min, x_max, image_width)
                    height_condition = step['condition'](y_min, y_max, image_height)
                    if width_condition or height_condition:
                        size_message = ['width'] if width_condition else []
                        size_message.extend(['height'] if height_condition else [])
                        message = step['message'].format(annotation.file_path, index, ' ', ' and '.join(size_message),
                                                         region.points_x, region.points_y)

                        step['choice'] = input_feedback(message, step['choice'], step['choices'])

                        choice_op = step['choice'].lower()
                        # if skip the shapes
                        if choice_op == 's':
                            delete_regions[index] = True
                            message = step['message'].format(annotation.file_path, index,
                                                             '{} '.format(step['choices'][choice_op]),
                                                             ' and '.join(size_message),
                                                             region.points_x, region.points_y)
                            logger.info(message)

                            break
                        else:
                            region.points_x = list(map(partial(step['transform'], size=image_width), region.points_x))
                            region.points_y = list(map(partial(step['transform'], size=image_height), region.points_y))

                            message = step['message'].format(annotation.file_path, index,
                                                             '{} '.format(step['choices'][choice_op]),
                                                             ' and '.join(size_message),
                                                             region.points_x, region.points_y)
                            logger.info(message)

            # delete regions after iteration is finished
            for index in sorted(list(delete_regions.keys()), reverse=True):
                del annotation.regions[index]

        print('Finished convert image annotations from {}'.format(self.input_adapter.path))

    def _save_semantic_masks(self, annotation_keys):
        """
        Create semantic segmentation mask png files out of the annotations.
        The mask file name is the same as the image file name but is stored in png format.
        `annotation_keys`: The annotation keys to create the segmentation masks for
        """
        from skimage import draw

        num_masks = len(annotation_keys)
        logger.info('Start create {} segmentation masks in {}'.format(num_masks, self.semantic_mask_folder))

        # only the trainval images have annotation, not the test images
        for index, key in enumerate(annotation_keys):
            annotation = self.annotations[key]

            if not annotation.regions:
                continue

            image, image_width, image_height = get_image_size(annotation.file_path)

            # Convert polygons to a bitmap mask of shape
            # [height, width]
            mask = np.zeros((image_height, image_width), dtype=np.uint8)

            # sort the regions by category priority for handling pixels which are assigned to more than one category
            # the category with higher index paint over the category with lower index
            for region in sorted(annotation.regions, key=lambda r: self.categories.index(r.labels[0])):
                class_id = self.categories.index(region.labels[0]) + 1

                if region.shape == RegionShape.NONE:
                    continue
                if region.shape == RegionShape.RECTANGLE:
                    rr, cc = draw.rectangle((region.points_y[0], region.points_x[0]),
                                            (region.points_y[1], region.points_x[1]))
                elif region.shape == RegionShape.POLYGON:
                    rr, cc = draw.polygon(region.points_y, region.points_x)
                else:
                    # handle POINT, CIRCLE and ELLIPSE
                    r_x, r_y = (1, 1) if region.shape == RegionShape.POINT else (region.radius_x, region.radius_y)
                    rr, cc = draw.ellipse(region.points_y[0], region.points_x[0], r_y, r_x)

                # Get indexes of pixels inside the polygon and set them to 1
                mask[rr, cc] = class_id

            # save the semantic mask
            file_name = basename(annotation.file_path)
            mask_path = join(self.semantic_mask_folder, splitext(file_name)[0] + '.png')
            write_mask(mask, mask_path, self.palette)

            logger.info('{} / {} - Created segmentation mask {}'.format(index + 1, num_masks, mask_path))

        logger.info('Finish create {} segmentation masks in {}'.format(num_masks, self.semantic_mask_folder))



# AUTOGENERATED! DO NOT EDIT! File to edit: tensorflow-tflite_metadata.ipynb (unless otherwise specified).

__all__ = ['SAVED_MODEL_META_DEFAULT_KEY', 'AUTHOR', 'logger', 'MetaInfo', 'create_metadata', 'write_metadata',
           'read_metadata', 'configure_logging']


# Cell
import logging
import logging.handlers
import argparse
import sys
import tensorflow as tf
from tflite_support import flatbuffers
from tflite_support import metadata as _metadata
from tflite_support import metadata_schema_py_generated as _metadata_fb
from os.path import basename
from ..dataset.type import DatasetType, infer_dataset_type


# Cell
SAVED_MODEL_META_DEFAULT_KEY = 'serving_default'
AUTHOR = 'Protosolution'


# Cell
logger = logging.getLogger(__name__)


# Cell
class MetaInfo:
    """
    Metadata information.
    `name`: The metadata name.
    `desc`: The metadata description.
    `prop`: The metadata property.
    `prop_type`: The metadata property type.
    `range_min`: The metadata min range.
    `range_max`: The metadata max range.
    `stats_min`: A list of min statistics per channel.
    `stats_max`: A list of max statistics per channel.
    `associated_files`: A list of associated files.
    """

    def __init__(self, name=None, desc=None, prop=None, prop_type=None, range_min=None, range_max=None,
                 stats_min=None, stats_max=None, associated_files=None):
        self.name = name
        self.desc = desc
        self.prop = prop
        self.prop_type = prop_type
        self.range_min = range_min
        self.range_max = range_max
        self.stats_min = stats_min
        self.stats_max = stats_max
        self.associated_files = associated_files

    def has_range(self):
        return self.range_min is not None or self.range_max is not None

    def has_stats(self):
        return self.stats_min or self.stats_max


# Cell
def create_metadata(saved_model_dir, categories_file_path, model_type, model_name, model_version=1):
    """
    Write metadata to the Tensowflow Lite model on disk.
    `saved_model_dir`: the path to the folder containing the SavedModel
    `categories_file_path`: the path to the categories.txt file
    `model_type`: the type of the model
    `model_name`: the name of the model
    `model_version`: the version of the model
    returns: the model metadata
    """
    saved_model = tf.saved_model.load(saved_model_dir)
    saved_model_meta = saved_model.signatures[SAVED_MODEL_META_DEFAULT_KEY]

    # Creates model info.
    model_meta = _metadata_fb.ModelMetadataT()
    model_meta.name = model_name

    description = ""
    if model_type == DatasetType.IMAGE_OBJECT_DETECTION:
        description = ("Identify which of a known set of objects "
                       "might be present and provide information about their positions "
                       "within the given image or a video stream.")
    elif model_type == DatasetType.IMAGE_CLASSIFICATION:
        description = ("Identify the most prominent object in the "
                       "image from a set of ategories.")

    model_meta.description = description
    model_meta.version = "v{}".format(model_version)
    model_meta.author = AUTHOR
    model_meta.license = ("Apache License. Version 2.0 "
                          "http://www.apache.org/licenses/LICENSE-2.0.")

    # Creates input info.
    input_meta = _create_input_metadata(saved_model_meta)

    # Creates output info.
    output_meta, output_groups = _create_output_metadata(saved_model_meta, categories_file_path, model_type)

    # Creates subgraph info.
    subgraph = _metadata_fb.SubGraphMetadataT()
    subgraph.inputTensorMetadata = input_meta
    subgraph.outputTensorMetadata = output_meta
    subgraph.outputTensorGroups = output_groups
    model_meta.subgraphMetadata = [subgraph]

    return model_meta


# Cell
def write_metadata(model_meta, model_path, categories_file_path):
    """
    Write metadata to the Tensowflow Lite model on disk.
    `model_meta`: the model metadata
    `model_path`: the path to the Tensorflow Lite model
    `categories_file_path`: the path to the categories.txt file
    """

    b = flatbuffers.Builder(0)
    b.Finish(
        model_meta.Pack(b),
        _metadata.MetadataPopulator.METADATA_FILE_IDENTIFIER)
    metadata_buf = b.Output()

    populator = _metadata.MetadataPopulator.with_model_file(model_path)
    populator.load_metadata_buffer(metadata_buf)
    populator.load_associated_files([categories_file_path])
    populator.populate()


# Cell
def read_metadata(model_path):
    """
    Read meta-data from the Tensowflow Lite model on disk.
    `model_path`: the path to the Tensorflow Lite model
    returns: the metadata in JSON format
    """
    displayer = _metadata.MetadataDisplayer.with_model_file(model_path)
    metadata = displayer.get_metadata_json()
    return metadata


# Cell
def _create_input_metadata(saved_model_meta, input_min=0, input_max=255, norm_mean=127.5, norm_std=127.5):
    """
    Creates input metadata

    `saved_model_meta`: The saved model meta data.
    `input_min`: The input min value.
    `input_max`: The input max value.
    `norm_mean`: The normalization mean value.
    `norm_std`: The normalization std value.
    """
    model_input = saved_model_meta.inputs[0]
    _, width, height, channel = model_input.shape

    if channel == 1:
        color_space = _metadata_fb.ColorSpaceType.GRAYSCALE
        channel_description = "one channel (grayscale)"
    if channel == 3:
        color_space = _metadata_fb.ColorSpaceType.RGB
        channel_description = "three channels (red, blue, and green)"
    else:
        color_space = _metadata_fb.ColorSpaceType.UNKNOWN
        channel_description = "{} channel".format(channel)

    # Creates input info.
    desc = (
        "Input image. The expected image is {0} x {1}, with "
        "{2} per pixel. Each value in the tensor "
        "is between {3} and {4}.".format(width, height, channel_description, input_min, input_max))
    image_prop = _metadata_fb.ImagePropertiesT()
    image_prop.colorSpace = color_space

    input_normalization = _metadata_fb.ProcessUnitT()
    input_normalization.optionsType = (
        _metadata_fb.ProcessUnitOptions.NormalizationOptions)
    input_normalization.options = _metadata_fb.NormalizationOptionsT()
    input_normalization.options.mean = [norm_mean]
    input_normalization.options.std = [norm_std]

    meta_info = MetaInfo(name="image",
                         desc=desc,
                         prop=image_prop,
                         prop_type=_metadata_fb.ContentProperties.ImageProperties,
                         stats_min=[input_min],
                         stats_max=[input_max])

    input_meta = _create_tensor_metadata(meta_info)
    input_meta.processUnits = [input_normalization]

    return [input_meta]


# Cell
def _create_output_metadata(saved_model_meta, categories_file_path, model_type):
    """
    Creates output metadata

    `saved_model_meta`: The saved model meta data.
    `categories_file_path`: the path to the categories.txt file.
    `model_type`: The type of the model.
    """
    output_meta = []
    output_groups = []

    if model_type == Type.IMAGE_OBJECT_DETECTION:
        output_meta = [
            MetaInfo(name="location",
                     desc="The locations of the detected boxes.",
                     prop=_create_bbox_content_property_metadata(),
                     prop_type=_metadata_fb.ContentProperties.BoundingBoxProperties,
                     range_min=2,
                     range_max=2),
            MetaInfo(name="category",
                     desc="The categories of the detected boxes.",
                     prop=_metadata_fb.FeaturePropertiesT(),
                     prop_type=_metadata_fb.ContentProperties.FeatureProperties,
                     range_min=2,
                     range_max=2,
                     associated_files=[
                         _create_associated_files_metadata(categories_file_path,
                                                           "Labels for objects that the model can recognize.",
                                                           _metadata_fb.AssociatedFileType.TENSOR_VALUE_LABELS)
                     ]),
            MetaInfo(name="score",
                     desc="The scores of the detected boxes.",
                     prop=_metadata_fb.FeaturePropertiesT(),
                     prop_type=_metadata_fb.ContentProperties.FeatureProperties,
                     range_min=2,
                     range_max=2),
            MetaInfo(name="number of detections",
                     desc="The number of the detected boxes.",
                     prop=_metadata_fb.FeaturePropertiesT(),
                     prop_type=_metadata_fb.ContentProperties.FeatureProperties),
        ]
        output_groups = _metadata_fb.TensorGroupT()
        output_groups.name = "detection result"
        output_groups.tensorNames = [o.name for o in output_meta[:3]]
        output_groups = [output_groups]

    elif model_type == Type.IMAGE_CLASSIFICATION:
        output_meta = [
            MetaInfo(name="probability",
                     desc="Probabilities of the labels respectively.",
                     prop=_metadata_fb.FeaturePropertiesT(),
                     prop_type=_metadata_fb.ContentProperties.FeatureProperties,
                     stats_min=[0.0],
                     stats_max=[1.0],
                     associated_files=[
                         _create_associated_files_metadata(categories_file_path,
                                                           "Labels for objects that the model can recognize.")
                     ]),
        ]

    if output_meta:
        output_meta = [_create_tensor_metadata(m) for m in output_meta]

    return output_meta, output_groups


# Cell
def _create_tensor_metadata(meta_info: MetaInfo):
    """
    Creates tensor metadata

    `meta_info`: The metadata information to create a tensor metadata for
    returns: The tensor metadata.
    """
    meta = _metadata_fb.TensorMetadataT()
    meta.name = meta_info.name
    meta.description = meta_info.desc

    meta.content = _metadata_fb.ContentT()
    meta.content.content_properties = meta_info.prop
    meta.content.contentPropertiesType = meta_info.prop_type
    meta.associatedFiles = meta_info.associated_files
    if meta_info.has_range:
        meta.content.range = _metadata_fb.ValueRangeT()
        meta.content.range.min = 0 if meta_info.range_min is None else meta_info.range_min
        meta.content.range.max = 0 if meta_info.range_max is None else meta_info.range_max

    if meta_info.has_stats:
        meta.stats = _metadata_fb.StatsT()
        meta.stats.max = meta_info.stats_max
        meta.stats.min = meta_info.stats_min

    return meta


# Cell
def _create_bbox_content_property_metadata(bbox_type=None, bbox_index=None):
    """
    Creates bounding box property content metadata
    `bbox_type`: The bounding box type.
    `bbox_index`: The bounding box index.
    returns: The bounding box property content.
    """
    properties = _metadata_fb.BoundingBoxPropertiesT()
    properties.index = [1, 0, 3, 2] if bbox_index is None else bbox_index
    properties.type = _metadata_fb.BoundingBoxType.BOUNDARIES if bbox_type is None else bbox_type
    return properties


# Cell
def _create_associated_files_metadata(categories_file_path, desc, label_type=None):
    """
    Creates associated files metadata
    `categories_file_path`: the path to the categories.txt file
    `desc`: The tensor metadata description.
    `label_type`: The label type.
    returns: The associated files metadata.
    """
    associated_file = _metadata_fb.AssociatedFileT()
    associated_file.name = basename(categories_file_path)
    associated_file.description = desc
    associated_file.type = _metadata_fb.AssociatedFileType.TENSOR_AXIS_LABELS if label_type is None else label_type
    return associated_file


# Cell
def configure_logging(logging_level=logging.INFO):
    """
    Configures logging for the system.

    `logging_level`: The logging level to use.
    """
    logger.setLevel(logging_level)

    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging_level)

    logger.addHandler(handler)


# Cell
if __name__ == '__main__' and '__file__' in globals():
    configure_logging()

    parser = argparse.ArgumentParser()
    parser.add_argument("model",
                        help="The path to the Tensorflow Lite exported model file.")
    parser.add_argument("-s",
                        "--source",
                        help="The path to the folder containing the SavedModel.",
                        type=str,
                        default=None)
    parser.add_argument("-c",
                        "--categories",
                        help="The categories file to add to the Tensorflow Lite model.",
                        type=str,
                        default=None)
    parser.add_argument("-n",
                        "--name",
                        help="The name of the model.",
                        type=str,
                        default=None)
    parser.add_argument("-v",
                        "--version",
                        help="The version of the model.",
                        type=int,
                        default=1)
    parser.add_argument("-t",
                        "--type",
                        help="The type of the model, if not explicitly set try to infer from categories file path.",
                        choices=list(DatasetType),
                        type=DatasetType,
                        default=None)

    args = parser.parse_args()

    if args.categories is None:
        metadata = read_metadata(args.model)
        logger.info('Read metadata from Tensorflow Lite model: {}'.format(args.model))
        logger.info(metadata)
    else:
        model_type = args.type

        # try to infer the model type if not explicitly set
        if model_type is None:
            try:
                model_type = infer_dataset_type(args.categories)
            except ValueError as e:
                logger.error(e)
                sys.exit(1)

        model_meta = create_metadata(args.source, args.categories, model_type, args.name, args.version)
        write_metadata(model_meta, args.model, args.categories)

    logger.info('FINISHED!!!')



# AUTOGENERATED! DO NOT EDIT! File to edit: annotation-via_adapter.ipynb (unless otherwise specified).

__all__ = ['DEFAULT_ANNOTATIONS_FILE', 'DEFAULT_CATEGORY_LABEL_KEY', 'CSV_FIELDNAMES', 'logger', 'VIAAnnotationAdapter']


# Cell
import json
import csv
import shutil
import logging
from os.path import join, splitext, getsize, basename, isfile
from ..core import assign_arg_prefix
from .core import Annotation, AnnotationAdapter, Region, RegionShape, parse_region_shape, SubsetType
from ..io.core import create_folder


# Cell
DEFAULT_ANNOTATIONS_FILE = 'via_region_data.json'
DEFAULT_CATEGORY_LABEL_KEY = 'category'
CSV_FIELDNAMES = ['#filename', 'file_size', 'file_attributes', 'region_count', 'region_id', 'region_shape_attributes',
                  'region_attributes']


# Cell
logger = logging.getLogger(__name__)


# Cell
class VIAAnnotationAdapter(AnnotationAdapter):
    """
    Adapter to read and write annotations in the VIA annotation.
    """

    def __init__(self, path, categories_file_name=None, annotations_file_name=None, category_label_key=None):
        """
        VIA Adapter to read and write annotations.
        `path`: the folder containing the annotations
        `categories_file_name`: the name of the categories file
        `annotations_file_name`: the name of annotations file
        `category_label_key`: the key of the category label
        """
        super().__init__(path, categories_file_name)
        if annotations_file_name is None:
            self.annotations_file_name = DEFAULT_ANNOTATIONS_FILE
        else:
            self.annotations_file_name = annotations_file_name

        if category_label_key is None:
            self.category_label_key = DEFAULT_CATEGORY_LABEL_KEY
        else:
            self.category_label_key = category_label_key

    @classmethod
    def argparse(cls, prefix=None):
        """
        Returns the argument parser containing argument definition for command line use.
        `prefix`: a parameter prefix to set, if needed
        return: the argument parser
        """
        parser = super(VIAAnnotationAdapter, cls).argparse(prefix=prefix)
        parser.add_argument(assign_arg_prefix('--annotations_file_name', prefix),
                            dest="annotations_file_name",
                            help="The name of annotations file.",
                            default=None)
        parser.add_argument(assign_arg_prefix('--category_label_key', prefix),
                            dest="category_label_key",
                            help="The key of the category label.",
                            default=None)

        return parser

    def read_annotations(self, subset_type=SubsetType.TRAINVAL):
        """
        Reads a VIA annotations file.
        Supports JSON and CSV file format.
        Format is inferred from the annotations_file setting.
        `subset_type`: the subset type to read
        return: the annotations as dictionary
        """
        path = join(self.path, str(subset_type))
        annotations_file_name = self._annotation_file_name_suffix_handling(subset_type)
        annotations_file_path = join(self.path, annotations_file_name)
        logger.info('Read file sources from {}'.format(path))
        logger.info('Read annotations from {}'.format(annotations_file_path))
        return self._read_annotations_v1(path, annotations_file_path)

    def _read_annotations_v1(self, path, annotations_file_path):
        """
        Reads a VIA v1 annotations file.
        Supports JSON and CSV file format.
        `path`: the path to read file sources from
        `annotation_file_path`: the path to the annotations file
        return: the annotations as dictionary
        """
        file_extension = splitext(annotations_file_path)[1]

        if file_extension.lower() == '.json':
            logger.info('Read VIA v1 annotations in JSON format')
            annotations = self._read_annotations_v1_json(path, annotations_file_path)
        elif file_extension.lower() == '.csv':
            logger.info('Read VIA v1 annotations in CSV format')
            annotations = self._read_annotations_v1_csv(path, annotations_file_path)
        else:
            message = 'Unsupported annotation format at {}'.format(annotations_file_path)
            logger.error(message)
            raise ValueError(message)

        return annotations

    def _read_annotations_v1_csv(self, path, annotations_file_path):
        """
        Reads a VIA v1 CSV annotations file.
        `path`: the path to read file sources from
        `annotation_file_path`: the path to the annotations file
        return: the annotations as dictionary
        """
        annotations = {}

        with open(annotations_file_path, newline='') as csv_file:
            reader = csv.DictReader(csv_file)

            skipped_annotations = []
            for row in reader:
                file_path = join(path, row['#filename'])
                if not isfile(file_path):
                    logger.warning("{}: Source file not found, skip annotation.".format(file_path))
                    skipped_annotations.append(file_path)
                    continue

                annotation_id = "{}{}".format(row['#filename'], row['file_size'])

                if annotation_id not in annotations:
                    annotations[annotation_id] = Annotation(annotation_id=annotation_id, file_path=file_path)

                annotation = annotations[annotation_id]

                region_shape_attributes = json.loads(row['region_shape_attributes'])
                region = self._parse_region_shape_attributes(region_shape_attributes)
                region_attributes = json.loads(row['region_attributes'])
                category = None
                if region_attributes and self.category_label_key in region_attributes:
                    category = region_attributes[self.category_label_key]
                region.labels = [category] if category else []
                annotation.regions.append(region)

        logger.info('Finished read annotations')
        logger.info('Annotations read: {}'.format(len(annotations)))
        if skipped_annotations:
            logger.info('Annotations skipped: {}'.format(len(skipped_annotations)))

        return annotations

    def _read_annotations_v1_json(self, path, annotations_file_path):
        """
        Reads a VIA v1 JSON annotations file.
        `path`: the path to read file sources from
        `annotation_file_path`: the path to the annotations file
        return: the annotations as dictionary
        """
        annotations = {}

        with open(annotations_file_path) as json_file:
            via_annotations = json.load(json_file)

            skipped_annotations = []
            for data in via_annotations.values():
                file_path = join(path, data['filename'])
                if not isfile(file_path):
                    logger.warning("{}: Source file not found, skip annotation.".format(file_path))
                    skipped_annotations.append(file_path)
                    continue

                annotation_id = "{}{}".format(data['filename'], data['size'])

                if annotation_id not in annotations:
                    annotations[annotation_id] = Annotation(annotation_id=annotation_id, file_path=file_path)

                annotation = annotations[annotation_id]

                for region_data in data['regions'].values():
                    region_shape_attributes = region_data['shape_attributes']
                    region = self._parse_region_shape_attributes(region_shape_attributes)
                    region_attributes = region_data['region_attributes']
                    category = None
                    if region_attributes and self.category_label_key in region_attributes:
                        category = region_attributes[self.category_label_key]
                    region.labels = [category] if category else []
                    annotation.regions.append(region)

        logger.info('Finished read annotations')
        logger.info('Annotations read: {}'.format(len(annotations)))
        if skipped_annotations:
            logger.info('Annotations skipped: {}'.format(len(skipped_annotations)))

        return annotations

    def write_annotations(self, annotations, subset_type=SubsetType.TRAINVAL):
        """
        Writes a VIA annotations file and copy the corresponding source files.
        Supports JSON and CSV file format.
        Format is inferred from the annotations_file setting.
        `annotations`: the annotations as dictionary
        `subset_type`: the subset type to write
        return: a list of written target file paths
        """
        path = join(self.path, str(subset_type))
        target_folder = create_folder(path)
        annotations_file_name = self._annotation_file_name_suffix_handling(subset_type)
        annotations_file_path = join(self.path, annotations_file_name)
        logger.info('Write file sources to {}'.format(target_folder))
        logger.info('Write annotations to {}'.format(annotations_file_path))

        return self._write_annotations_v1(path, annotations_file_path, annotations)

    def _write_annotations_v1(self, path, annotations_file_path, annotations):
        """
        Writes a VIA v1 annotations file and copy the corresponding source files.
        Supports JSON and CSV file format.
        `path`: the path to write file sources into
        `annotation_file_path`: the path to the annotations file
        `annotations`: the annotations to write
        return: a list of written source file paths
        """
        file_extension = splitext(annotations_file_path)[1]

        if file_extension.lower() == '.json':
            logger.info('Write VIA v1 annotations in JSON format')
            copied_files = self._write_annotations_v1_json(path, annotations_file_path, annotations)
        elif file_extension.lower() == '.csv':
            logger.info('Write VIA v1 annotations in CSV format')
            copied_files = self._write_annotations_v1_csv(path, annotations_file_path, annotations)
        else:
            message = 'Unsupported annotation format at {}'.format(annotations_file_path)
            logger.error(message)
            raise ValueError(message)
        return copied_files

    def _write_annotations_v1_csv(self, path, annotations_file_path, annotations):
        """
        Writes a VIA v1 CSV annotations file and copy the corresponding source files.
        `path`: the path to write file sources into
        `annotation_file_path`: the path to the annotations file
        `annotations`: the annotations to write
        return: a list of written source file paths
        """
        with open(annotations_file_path, 'w', newline='') as csv_file:
            writer = csv.DictWriter(csv_file, fieldnames=CSV_FIELDNAMES)
            writer.writeheader()

            skipped_annotations = []
            copied_files = []
            for annotation in annotations.values():
                target_file = join(path, basename(annotation.file_path))

                if not isfile(annotation.file_path):
                    logger.warning("{}: Source file not found, skip annotation.".format(annotation.file_path))
                    skipped_annotations.append(annotation.file_path)
                    continue
                if isfile(target_file):
                    logger.warning("{}: Target file already exist, skip annotation.".format(annotation.file_path))
                    skipped_annotations.append(annotation.file_path)
                    continue

                file_size = getsize(annotation.file_path)
                file_name = basename(annotation.file_path)
                for index, region in enumerate(annotation.regions):
                    region_shape_attributes = self._create_region_shape_attributes(region)
                    region_attributes = {
                        self.category_label_key: ' '.join(region.labels) if len(region.labels) else ''
                    }

                    writer.writerow({'#filename': file_name,
                                     'file_size': file_size,
                                     'file_attributes': '{}',
                                     'region_count': len(annotation.regions),
                                     'region_id': str(index),
                                     'region_shape_attributes': json.dumps(region_shape_attributes),
                                     'region_attributes': json.dumps(region_attributes)})
                # copy the file
                shutil.copy2(annotation.file_path, target_file)
                copied_files.append(target_file)

        logger.info('Finished write annotations')
        logger.info('Annotations written: {}'.format(len(annotations) - len(skipped_annotations)))
        if skipped_annotations:
            logger.info('Annotations skipped: {}'.format(len(skipped_annotations)))
        return copied_files

    def _write_annotations_v1_json(self, path, annotations_file_path, annotations):
        """
        Writes a VIA v1 JSON annotations file and copy the corresponding source files.
        `path`: the path to write file sources into
        `annotation_file_path`: the path to the annotations file
        `annotations`: the annotations to write
        return: a list of written source file paths
        """
        json_annotations = {}
        skipped_annotations = []
        copied_files = []

        for annotation in annotations.values():
            target_file = join(path, basename(annotation.file_path))

            if not isfile(annotation.file_path):
                logger.warning("{}: Source file not found, skip annotation.".format(annotation.file_path))
                skipped_annotations.append(annotation.file_path)
                continue
            if isfile(target_file):
                logger.warning("{}: Target file already exist, skip annotation.".format(annotation.file_path))
                skipped_annotations.append(annotation.file_path)
                continue

            file_size = getsize(annotation.file_path)
            file_name = basename(annotation.file_path)
            file_id = '{:s}{:d}'.format(file_name, file_size)
            regions = {}
            for index, region in enumerate(annotation.regions):
                regions[str(index)] = {
                    'shape_attributes': self._create_region_shape_attributes(region),
                    'region_attributes': {
                        self.category_label_key: ' '.join(region.labels) if len(region.labels) else ''
                    }
                }
            json_annotations[file_id] = {
                'fileref': "",
                'size': file_size,
                'filename': file_name,
                'base64_img_data': "",
                'file_attributes': '{}',
                "regions": regions
            }
            # copy the file
            shutil.copy2(annotation.file_path, target_file)
            copied_files.append(target_file)

        with open(annotations_file_path, 'w') as json_file:
            json.dump(json_annotations, json_file)

        logger.info('Finished write annotations')
        logger.info('Annotations written: {}'.format(len(annotations) - len(skipped_annotations)))
        if skipped_annotations:
            logger.info('Annotations skipped: {}'.format(len(skipped_annotations)))
        return copied_files

    def _annotation_file_name_suffix_handling(self, subset_type):
        """
        Handle annotations file name based on the subset type.
        `subset_type`: the subset type to handle
        return: the annotations file name
        """
        file_name, ext = splitext(self.annotations_file_name)
        if subset_type in [SubsetType.TRAIN, SubsetType.VAL] and not file_name.endswith(str(subset_type)):
            suffix = "_{}".format(str(subset_type))
            return "{}{}{}".format(file_name, suffix, ext)
        return self.annotations_file_name

    @classmethod
    def _parse_region_shape_attributes(cls, region_shape_attributes):
        """
        Parse region shape attributes.
        `region_shape_attributes`: the region shape attributes as dictionary
        return: the corresponding annotation
        """
        if not region_shape_attributes:
            return Annotation()

        region_shape = parse_region_shape(region_shape_attributes['name'])
        points_x = None
        points_y = None
        radius_x = 0
        radius_y = 0
        if region_shape == RegionShape.CIRCLE:
            points_x = [region_shape_attributes['cx']]
            points_y = [region_shape_attributes['cy']]
            radius_x = region_shape_attributes['r']
            radius_y = region_shape_attributes['r']
        elif region_shape == RegionShape.ELLIPSE:
            points_x = [region_shape_attributes['cx']]
            points_y = [region_shape_attributes['cy']]
            radius_x = region_shape_attributes['rx']
            radius_y = region_shape_attributes['ry']
        elif region_shape == RegionShape.POINT:
            points_x = [region_shape_attributes['cx']]
            points_y = [region_shape_attributes['cy']]
        elif region_shape == RegionShape.POLYGON:
            points_x = region_shape_attributes['all_points_x']
            points_y = region_shape_attributes['all_points_y']
        elif region_shape == RegionShape.RECTANGLE:
            x = region_shape_attributes['x']
            y = region_shape_attributes['y']
            width = region_shape_attributes['width']
            height = region_shape_attributes['height']
            points_x = [x, x + width]
            points_y = [y, y + height]
        return Region(shape=region_shape, points_x=points_x, points_y=points_y, radius_x=radius_x, radius_y=radius_y)

    @classmethod
    def _create_region_shape_attributes(cls, region: Region):
        """
        Create region shape attributes.
        `region`: the region to create region shape attributes from
        return: the corresponding region shape attributes as dictionary
        """
        region_shape_attributes = {
            "name": str(region.shape),

        }
        c_x = region.points_x[0] if len(region.points_x) else 0
        c_y = region.points_y[0] if len(region.points_y) else 0

        if region.shape == RegionShape.CIRCLE:
            region_shape_attributes['cx'] = c_x
            region_shape_attributes['cy'] = c_y
            region_shape_attributes['r'] = max(region.radius_x, region.radius_y)
        elif region.shape == RegionShape.ELLIPSE:
            region_shape_attributes['cx'] = c_x
            region_shape_attributes['cy'] = c_y
            region_shape_attributes['rx'] = region.radius_x
            region_shape_attributes['ry'] = region.radius_y
        elif region.shape == RegionShape.POINT:
            region_shape_attributes['cx'] = c_x
            region_shape_attributes['cy'] = c_y
        elif region.shape == RegionShape.POLYGON:
            region_shape_attributes['all_points_x'] = region.points_x
            region_shape_attributes['all_points_y'] = region.points_y
        elif region.shape == RegionShape.RECTANGLE:
            region_shape_attributes['x'] = region.points_x[0]
            region_shape_attributes['y'] = region.points_y[0]
            region_shape_attributes['width'] = region.points_x[1] - region.points_x[0]
            region_shape_attributes['height'] = region.points_y[1] - region.points_y[0]
        return region_shape_attributes



# AUTOGENERATED! DO NOT EDIT! File to edit: evaluation-core.ipynb (unless otherwise specified).

__all__ = ['logger', 'box_area', 'intersection_box', 'union_box', 'intersection_over_union', 'configure_logging']


# Cell
import argparse
import logging
import sys


# Cell
logger = logging.getLogger(__name__)


# Cell
def box_area(box):
    """
    Calculates the area of a bounding box.
    Source code mainly taken from:
    https://www.pyimagesearch.com/2016/11/07/intersection-over-union-iou-for-object-detection/
    `box`: the bounding box to calculate the area for with the format ((x_min, x_max), (y_min, y_max))
    return: the bounding box area
    """
    return max(0, box[0][1] - box[0][0] + 1) * max(0, box[1][1] - box[1][0] + 1)


# Cell
def intersection_box(box_a, box_b):
    """
    Calculates the intersection box from two bounding boxes with the format ((x_min, x_max), (y_min, y_max)).
    Source code mainly taken from:
    https://www.pyimagesearch.com/2016/11/07/intersection-over-union-iou-for-object-detection/
    `box_a`: the first box
    `box_b`: the second box
    return: the intersection box
    """
    # determine the (x, y)-coordinates of the intersection rectangle
    x_a = max(box_a[0][0], box_b[0][0])
    y_a = max(box_a[1][0], box_b[1][0])
    x_b = min(box_a[0][1], box_b[0][1])
    y_b = min(box_a[1][1], box_b[1][1])
    return (x_a, x_b), (y_a, y_b)


# Cell
def union_box(box_a, box_b):
    """
    Calculates the union box from two bounding boxes with the format ((x_min, x_max), (y_min, y_max)).
    Source code mainly taken from:
    https://www.pyimagesearch.com/2016/11/07/intersection-over-union-iou-for-object-detection/
    `box_a`: the first box
    `box_b`: the second box
    return: the union box
    """
    # determine the (x, y)-coordinates of the intersection rectangle
    x_a = min(box_a[0][0], box_b[0][0])
    y_a = min(box_a[1][0], box_b[1][0])
    x_b = max(box_a[0][1], box_b[0][1])
    y_b = max(box_a[1][1], box_b[1][1])
    return (x_a, x_b), (y_a, y_b)


# Cell
def intersection_over_union(box_a, box_b):
    """
    Intersection over Union (IoU) algorithm.
    Calculates the IoU from two bounding boxes with the format ((x_min, x_max), (y_min, y_max)).
    Source code mainly taken from:
    https://www.pyimagesearch.com/2016/11/07/intersection-over-union-iou-for-object-detection/
    `box_a`: the first box
    `box_b`: the second box
    return: the IoU
    """
    # determine the (x, y)-coordinates of the intersection rectangle
    inter_box = intersection_box(box_a, box_b)
    # compute the area of intersection rectangle
    inter_area = box_area(inter_box)
    # compute the area of both the prediction and ground-truth
    # rectangles
    box_a_area = box_area(box_a)
    box_b_area = box_area(box_b)
    # compute the intersection over union by taking the intersection
    # area and dividing it by the sum of prediction + ground-truth
    # areas - the interesection area
    iou = inter_area / float(box_a_area + box_b_area - inter_area)
    # return the intersection over union value
    return iou


# Cell
def configure_logging(logging_level=logging.INFO):
    """
    Configures logging for the system.

    :param logging_level: The logging level to use.
    """
    logger.setLevel(logging_level)

    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging_level)

    logger.addHandler(handler)


# Cell
if __name__ == '__main__' and '__file__' in globals():
    # for direct shell execution
    configure_logging()

    parser = argparse.ArgumentParser()
    parser.add_argument("annotation",
                        help="The path to the VIA annotation file.")

    args = parser.parse_args()
